/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include "elem-corba-any.h"


typedef void (*NxpElementToCorbaAnyFunc)(NxpElement *elem, CORBA_any *any, CORBA_Environment *ev);


/**
 * element_to_corba_any:
 * @elem: source #NxpElement
 * @any: pointer to CORBA_any to be filled with contents
 * 
 * Convert a numexp element into CORBA_any. Some common element types
 * are converted to similar CORBA types (int, string, real->double,
 * etc.). All other types are represented as plain Numexp::Element
 * element references.
 **/
void element_to_corba_any(NxpElement *elem, CORBA_any *any, CORBA_Environment *ev)
{
    NxpElementToCorbaAnyFunc caster;

    caster = nxp_element_type_get_data(NXP_ELEMENT_GET_TYPE_INFO(elem),
				       "to_corba_any");
    any->_type = NULL;
    any->_value = NULL;
    if (caster) {
	caster(elem, any, ev);
    } else {
	Numexp_Element *el;
	el = Numexp_Element__alloc();
	*el = element2ref(elem);
	any->_type = TC_Numexp_Element;
	any->_value = el;
	CORBA_any_set_release(any, CORBA_TRUE);
    }
}


static void int_to_corba_any(NxpElement *elem, CORBA_any *any, CORBA_Environment *ev)
{
    GError *err = NULL;

    CORBA_long *x = ORBit_small_alloc(TC_CORBA_long);
    *x = nxp_element_int_get_si(elem, &err);
    if (err) {
	gerror_to_exception(err, ev);
	return;
    }
    /* FIXME : how do you want to report the error ? */
    any->_type = TC_CORBA_long;
    any->_value = x;
    CORBA_any_set_release(any, CORBA_TRUE);
}

static void void_to_corba_any(NxpElement *elem, CORBA_any *any, CORBA_Environment *ev)
{
    any->_type  = TC_void;
    any->_value = NULL;
    CORBA_any_set_release(any, CORBA_FALSE);
}

static void real_to_corba_any(NxpElement *elem, CORBA_any *any, CORBA_Environment *ev)
{
    GError *err = NULL;
    CORBA_float *x = ORBit_small_alloc(TC_CORBA_float);
    *x = nxp_element_real_get_double(elem, &err);
    if (err) {
	gerror_to_exception(err, ev);
	return;
    }

    any->_type  = TC_CORBA_float;
    any->_value = x;
    CORBA_any_set_release(any, CORBA_TRUE);
}


static void vector_to_corba_any(NxpElement *elem, CORBA_any *any, CORBA_Environment *ev)
{
    NxpElementVector *vec = NXP_ELEMENT_VECTOR(elem);
    CORBA_sequence_CORBA_any *seq = CORBA_sequence_CORBA_any__alloc();
    int i;

    seq->_buffer = CORBA_sequence_CORBA_any_allocbuf(vec->size);
    seq->_maximum = 0;
    seq->_length = vec->size;
    CORBA_sequence_set_release(seq, TRUE);
    for (i = 0; i < vec->size; ++i)
	element_to_corba_any(NXP_ELEMENT(_element_vector_index(vec, i)), &seq->_buffer[i], ev);
    any->_type  = TC_CORBA_sequence_CORBA_any;
    any->_value = seq;
    CORBA_any_set_release(any, CORBA_TRUE);
}


static void string_to_corba_any(NxpElement *elem, CORBA_any *any, CORBA_Environment *ev)
{
    CORBA_string *x = ORBit_small_alloc(TC_CORBA_string);
    *x = CORBA_string_dup(NXP_ELEMENT_STRING(elem)->string);
    any->_type  = TC_CORBA_string;
    any->_value = x;
    CORBA_any_set_release(any, CORBA_TRUE);
}


void register_corba_any_casters(NumexpKernelClass *class)
{
    nxp_element_type_set_data(nxp_element_type_get_info(NXP_ELEMENT_TYPE_INT),
			      "to_corba_any", int_to_corba_any);
    nxp_element_type_set_data(nxp_element_type_get_info(NXP_ELEMENT_TYPE_REAL),
			      "to_corba_any", real_to_corba_any);
    nxp_element_type_set_data(nxp_element_type_get_info(NXP_ELEMENT_TYPE_VECTOR),
			      "to_corba_any", vector_to_corba_any);
    nxp_element_type_set_data(nxp_element_type_get_info(NXP_ELEMENT_TYPE_STRING),
			      "to_corba_any", string_to_corba_any);
    nxp_element_type_set_data(nxp_element_type_get_info(NXP_ELEMENT_TYPE_VOID),
			      "to_corba_any", void_to_corba_any);
}
