/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#define _GNU_SOURCE
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <glib.h>
#include "kernel/nxp_kernel.h"
#include "parser/parser.h"

#include "numexp-kernel.h"
#include "elem-corba-any.h"
#include "nxp-ticket-source.h"


#define PARENT_TYPE BONOBO_OBJECT_TYPE


static GObjectClass *parent_class;
static NxpTicketSource *elem_tksrc = NULL; /* ticket source for elements    */
static NxpTicketSource *def_tksrc = NULL;  /* ticket source for definitions */

static void                   numexp_kernel_class_init (NumexpKernelClass *class);
static void                   numexp_kernel_base_init  (NumexpKernelClass *class);
static Numexp_Client_ArgList* server_to_client_arglist (GSList            *args);
static gboolean               listeners_timeout_func   (gpointer           data);
static void                   output_handler           (const gchar       *str,
							NxpMessageType     msg_type,
							NxpContext        *context);



/* NxpElement <-> Numexp_Element conversion functions */
NxpElement* numexp_kernel_ref2element(Numexp_Element ref)
{
    return NXP_ELEMENT(nxp_ticket_source_lookup_ticket
		       (elem_tksrc, (guint32) ref));
}


Numexp_Element numexp_kernel_element2ref(NxpElement *elem)
{
    return (Numexp_Element)
	nxp_ticket_source_create_ticket(elem_tksrc, elem);
}


/* NxpDefinition <-> Numexp_Definition conversion functions */
NxpDefinition* numexp_kernel_ref2definition(Numexp_DefinitionHandle ref)
{
    return (NxpDefinition*) nxp_ticket_source_lookup_ticket
	(def_tksrc, (guint32) ref);
}


Numexp_DefinitionHandle numexp_kernel_definition2ref(NxpDefinition *def)
{
    return (Numexp_DefinitionHandle)
	nxp_ticket_source_create_ticket(def_tksrc, def);
}


static void numexp_kernel_finalize(GObject *object)
{
    NumexpKernel *self = NUMEXP_KERNEL(object);

    if (self->client != CORBA_OBJECT_NIL) {
	bonobo_object_release_unref((Bonobo_Unknown) self->client, NULL);
	self->client = CORBA_OBJECT_NIL;
    }
    g_source_destroy(self->listeners_timeout);
    nxp_eval_context_destroy(self->context);
    /* --- */
    if (parent_class->finalize)
	parent_class->finalize(object);
}


static void numexp_kernel_init(NumexpKernel *self)
{
    self->client = CORBA_OBJECT_NIL;
    self->context = nxp_eval_context_new(NULL);
    nxp_eval_context_set_user_data(self->context, "kernel", self);
    self->context->output_handler = output_handler;

    self->listeners_timeout = g_timeout_source_new(100);
    g_source_set_callback(self->listeners_timeout, listeners_timeout_func, self, NULL);
    g_source_attach(self->listeners_timeout, NULL);
    // g_timeout_add(100, listeners_timeout_func, self);
}


GType numexp_kernel_get_type(void)
{
    static GType type = 0;

    if (!type) {
	GTypeInfo info = {
	    sizeof(NumexpKernelClass), 
	    (GBaseInitFunc) numexp_kernel_base_init,
	    (GBaseFinalizeFunc) NULL,
	    (GClassInitFunc) numexp_kernel_class_init, 
	    NULL, /* class_finalize */
	    NULL, /* class_data */
	    sizeof(NumexpKernel),
	    0, /* n_preallocs */
	    (GInstanceInitFunc) numexp_kernel_init,
	    NULL
	};

	type = bonobo_type_unique(
	    PARENT_TYPE, 
	    POA_Numexp_Kernel__init, NULL, 
	    G_STRUCT_OFFSET(NumexpKernelClass, epv), 
	    &info, "NumexpKernel");
    }

    return type;
}


BonoboObject* numexp_kernel_new(void)
{
    BonoboObject *obj;

    obj = g_object_new(NUMEXP_KERNEL_TYPE, NULL);
    return obj;
}


__inline__ void gerror_to_exception(GError *err, CORBA_Environment *ev)
{
    Numexp_ParseError *parse_err;
    Numexp_EvalError *eval_err;

    if (err != NULL) {
	if (err->domain == nxp_error_domain) {
	    CORBA_string str;
	    eval_err = Numexp_EvalError__alloc();
	    str = CORBA_string_alloc(strlen(err->message) + 1);
	    strcpy(str, err->message);
	    eval_err->detail = str;
	    CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_EvalError,
				eval_err);
	} else if (err->domain == EXP_PARSE_ERROR_DOMAIN) {
	    CORBA_string str;
	    parse_err = Numexp_ParseError__alloc();
	    str = CORBA_string_alloc(strlen(err->message) + 1);
	    strcpy(str, err->message);
	    parse_err->detail = str;
	    CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_ParseError,
				parse_err);
	}
	else
	    CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
	g_clear_error(&err);
    }
}


static Numexp_Element
impl_Numexp_Kernel_parseString(PortableServer_Servant servant,
			       const CORBA_char * expression,
			       CORBA_Environment * ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    Numexp_Element retval;
    NxpElement *elem;
    GError *err = NULL;

    g_return_val_if_fail(expression, 0);
    elem = exp_parse_string(expression, self->context, &err);
#if 1
    {
	gchar *s1 = g_strescape(expression, NULL);
	g_message("Parsing string \"%s\"", s1);
	g_free(s1);
    }
#endif
    if (err) {
	gerror_to_exception(err, ev);
	return 0;
    }   
    retval = element2ref(elem);
    return retval;
}


static void
impl_Numexp_Kernel_elementUnref(PortableServer_Servant servant,
				Numexp_Element elem, CORBA_Environment * ev)
{
    g_return_if_fail(ref2element(elem));
    NXP_ELEMENT_UNREF(ref2element(elem));
    nxp_ticket_source_free_ticket(elem_tksrc, elem);
}


static void
impl_Numexp_Kernel_elementRef(PortableServer_Servant servant,
				Numexp_Element elem, CORBA_Environment * ev)
{
    NxpElement *el = ref2element(elem);
    g_return_if_fail(el);
    NXP_ELEMENT_REF(el);
    /* Increase the reference count of the ticket */
    element2ref(el);
}


static void output_handler(const gchar *str, NxpMessageType msg_type, NxpContext *context)
{
    NumexpKernel *self = NUMEXP_KERNEL(nxp_eval_context_get_user_data(context, "kernel"));
    Numexp_Client_MessageType msg_type_corba;

    if (self->client == CORBA_OBJECT_NIL)
	return;
    
    switch (msg_type)
    {
    case NXP_MESSAGE_PARSER_WARNING:
	msg_type_corba = Numexp_Client_MESSAGE_PARSER_WARNING;
	break;
    case NXP_MESSAGE_EVAL_WARNING:
	msg_type_corba = Numexp_Client_MESSAGE_PARSER_WARNING;
	break;
    case NXP_MESSAGE_PRINT:
    default:
	msg_type_corba = Numexp_Client_MESSAGE_PRINT;
	break;
    }
    Numexp_Client_print(self->client, str, msg_type_corba, NULL);
}


static Numexp_Element
impl_Numexp_Kernel_elementEval(PortableServer_Servant servant,
			       Numexp_Element elem, CORBA_Environment *ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    Numexp_Element retval;
    NxpElement *el, *el1;
    GError *err = NULL;
    el = ref2element(elem);
    g_return_val_if_fail(el, 0);
/*     nxp_set_output_handler(output_handler, self); /\* This is broken on */
/*                                                      multi-threaded */
/*                                                      servers *\/ */
    el1 = nxp_element_eval(el, self->context, &err);
    if (err) {
	gerror_to_exception(err, ev);
	return 0;
    }
    g_assert(el1 != NULL);
    retval = element2ref(el1);
    return retval;
}

static Numexp_Element
impl_Numexp_Kernel_elementSimplify(PortableServer_Servant servant,
				   Numexp_Element elem, CORBA_Environment *ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    Numexp_Element retval;
    NxpElement *el, *el1;
    GError *err = NULL;
    el = ref2element(elem);
    g_return_val_if_fail(el, 0);
/*     nxp_set_output_handler(output_handler, self); /\* This is broken on */
/*                                                      multi-threaded */
/*                                                      servers *\/ */
    el1 = nxp_element_simplify(el, self->context, &err);

/*	In simplification mode, some bad users can enter ::> 1/0    :-), so...

    if (err && err->domain != EXP_PARSE_ERROR_DOMAIN)
        g_warning("returning unexpected error (%s: %s) in elementSimplify",
                  g_quark_to_string(err->domain), err->message);*/

    if (err) {
	gerror_to_exception(err, ev);
	return 0;
    }   
    retval = element2ref(el1);
    return retval;
}

static CORBA_char*
impl_Numexp_Kernel_version(PortableServer_Servant servant,
			   CORBA_Environment * ev)
{
    CORBA_char *retval;

    retval = CORBA_string_dup(nxp_kernel_version());
    return retval;
}

static CORBA_char*
impl_Numexp_Kernel_elementStr(PortableServer_Servant servant,
			      Numexp_Element elem,
			      CORBA_Environment * ev)
{
    CORBA_char *retval;
    GString *s;
    NxpElement *el;
    el = ref2element(elem);
    if (!el) {
        g_warning("ref2element(%u) failed!", elem);
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_BadParam, NULL);
        return NULL;
    }
    s = nxp_element_str(el);
    retval = CORBA_string_dup(s->str);
    g_string_free(s, TRUE);
    return retval;
}


static CORBA_char*
impl_Numexp_Kernel_elementToMathML(PortableServer_Servant servant,
				   Numexp_Element elem,
				   CORBA_Environment * ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    CORBA_char *retval;
    GString *s;
    NxpElement *el, *list;
    NxpElement *mathml_call, *mathml_result;
    NxpIdentifier *id;
    GError *err = NULL;

    el = ref2element(elem);
    g_return_val_if_fail(el, NULL);

    id = nxp_identifier_new();
    nxp_identifier_append_separator(id);
    nxp_identifier_append_component(id, "mathml");
    list = nxp_element_list_new_simple(el);
    mathml_call = nxp_element_function_new(id, list, NULL);
    NXP_ELEMENT_UNREF(list);
    nxp_identifier_unref(id);

    mathml_result = nxp_element_eval(mathml_call, self->context, &err);

    /* if ::mathml is not defined, load the mathml module and try again */
    if (err && err->domain == nxp_error_domain &&
	err->code == NXP_ERROR_UNDEFINED)
    {
	nxp_module_load("mathml", self->context);
	g_clear_error(&err);
	mathml_result = nxp_element_eval(mathml_call, self->context, &err);
    }

    NXP_ELEMENT_UNREF(mathml_call); /* no longer needed */

    if (err) {
	gerror_to_exception(err, ev);
	NXP_ELEMENT_XUNREF(mathml_result);
	return NULL;
    }

    if (NXP_ELEMENT_GET_TYPE(mathml_result) == NXP_ELEMENT_TYPE_STRING)
	retval = CORBA_string_dup(NXP_ELEMENT_STRING(mathml_result)->string);
    else {
	s = nxp_element_str(mathml_result);
	retval = CORBA_string_dup(s->str);
	g_string_free(s, TRUE);
    }
    NXP_ELEMENT_UNREF(mathml_result);
    return retval;
}


static void impl_Numexp_Kernel_setClient(PortableServer_Servant servant,
					 const Numexp_Client client,
					 CORBA_Environment *ev)
{
    CORBA_Environment ev1;
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));

    CORBA_exception_init(&ev1);

    if (self->client != CORBA_OBJECT_NIL)
	bonobo_object_release_unref((Bonobo_Unknown) self->client, &ev1);
    CORBA_exception_free(&ev1);

    self->client = bonobo_object_dup_ref((Bonobo_Unknown) client, NULL);
}

static void _corba_ex_to_gerror(CORBA_Environment *ev, GError **error)
{
    if (ev->_major == CORBA_USER_EXCEPTION) {
	/* Handle the 3 user exception that
	 * Numexp::Client::evalFunction is allowed to raise. */
	if (CORBA_TypeCode_equal(ev->_any._type, TC_Numexp_EvalError, NULL)) {
	    /* Evaluation error */
	    Numexp_EvalError *err = (Numexp_EvalError *) ev->_any._value;
	    g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, "%s", err->detail);	    
	    return;
	} else if (CORBA_TypeCode_equal(ev->_any._type, TC_Numexp_ParseError, NULL)) {
	    /* Parsing error */
	    Numexp_ParseError *err = (Numexp_ParseError *) ev->_any._value;
	    g_set_error(error, EXP_PARSE_ERROR_DOMAIN, EXP_PARSE_ERROR, "%s", err->detail);
	    return;
	} else if (CORBA_TypeCode_equal(ev->_any._type, TC_Numexp_NameError, NULL)) {
	    /* Name error */
	    g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED, "undefined");
	    return;
	}
    }
    /* Fallback error message */
    g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR,
		"Unknown exception when invoking client function: %s",
		bonobo_exception_get_text(ev));
}

static NxpElement* 
numexp_kernel_client_eval_func(NxpDefinition *definition, GSList *args, 
			       NxpEvalContext *context, GError **error)
{
    NumexpKernel *self = NUMEXP_KERNEL(nxp_eval_context_get_user_data(context, "kernel"));
    Numexp_Client_ArgList *arglist;
    Numexp_DefinitionHandle def;
    CORBA_Environment ev;
    Numexp_Element result;

    if (!self->client) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR,
		    "CORBA client is not available to evaluate function");
	return NULL;
    }
    CORBA_exception_init(&ev);
    def = nxp_ticket_source_lookup_data(def_tksrc, definition);
    arglist = server_to_client_arglist(args);
    result = Numexp_Client_evalFunction(self->client, def, arglist, &ev);
    CORBA_free(arglist);
    if (ev._major != CORBA_NO_EXCEPTION) {
	_corba_ex_to_gerror(&ev, error);
	return NULL;
    }
    CORBA_exception_free(&ev);
    if (result)
	return ref2element(result);
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, "Unknown CORBA error");	    
	return NULL;
    }
}


static Numexp_DefinitionHandle
impl_Numexp_Kernel_registerClientFunction(PortableServer_Servant servant,
					  const CORBA_char *name_with_path,
					  CORBA_Environment *ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    GError *error = NULL;
    NxpIdentifier *id;
    NxpDefinition *def;
    const char *def_name = NULL;
    NxpNameSpace *ns = NULL;
    Numexp_DefinitionHandle ret;

    g_message("impl_Numexp_Kernel_registerClientFunction('%s')", name_with_path);

    id = nxp_identifier_from_string(name_with_path, self->context, &error);
    if (error) {
	gerror_to_exception(error, ev);
	return 0;
    }
    nxp_identifier_resolve(id, self->context, &ns, &def_name);
    g_return_val_if_fail(ns, 0);
    def = nxp_name_space_lookup_definition_r(ns, def_name, FALSE);
    if (def) {
	NxpProtoInstance *inst = NULL;
	GSList *l;
	/* scan definition for a generic protype func(...)  */
	for (l = NXP_DEF_BI_FUNC(def)->prototypes; l; l = l->next) {
	    inst = (NxpProtoInstance *) l->data;
	    if (!inst->protoclass || inst->protoclass->arity == -1) break;
	}

	if (l && inst->eval_func == numexp_kernel_client_eval_func) {
	    g_message("defining corba client function: def exists with CORBA function");
	    nxp_identifier_unref(id);
	    return definition2ref(def);	/* function already defined by
                                           us: return existing handle */
	} else {
	    /* Existing definition will be replaced below. */
	}
    }
    def = nxp_bifunc_define_detailed(id, NULL,
				     numexp_kernel_client_eval_func,
				     NULL, numexp_kernel_client_eval_func,
				     NULL, NULL, self->context, 
				     NXP_PROTO_INSTANCE_DETAILED, 
				     &error);
    nxp_identifier_unref(id);
    if (def) {
	ret = definition2ref(def);
	nxp_definition_unref(def);
    } else
	ret = 0;
    if (error) {
	g_error("nxp_bifunc_define_detailed: %s", error->message);
	gerror_to_exception(error, ev);
	return 0;
    }   
    return ret;
}



static void numexp_kernel_base_init(NumexpKernelClass *class)
{
    nxp_kernel_init();
    register_corba_any_casters(class);
}



static void impl_Numexp_Kernel_functions_add_func(gpointer key, gpointer value, 
						  gpointer user_data) 
{
    Numexp_ListOfStrings* r;
    r = (Numexp_ListOfStrings*)user_data;
    
    r->_buffer[--r->_length] = CORBA_string_dup((char*)key);
}


static Numexp_ListOfStrings *
impl_Numexp_Kernel_functions(PortableServer_Servant servant,
			      CORBA_Environment * ev)
{


    NxpNameSpace *ns;
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    Numexp_ListOfStrings *retval = Numexp_ListOfStrings__alloc();

    ns = self->context->cwd;
    nxp_name_space_lock(ns);

    /* Initialize buffer size */
    retval->_length = g_hash_table_size(ns->definitions);
    retval->_buffer = Numexp_ListOfStrings_allocbuf(retval->_length);
    CORBA_sequence_set_release(retval, CORBA_TRUE);

    g_hash_table_foreach(ns->definitions,
			 impl_Numexp_Kernel_functions_add_func,
			 retval);
    // This is __NEEDED__ because I change it in the last function!!
    retval->_length = g_hash_table_size(ns->definitions);


    nxp_name_space_unlock(ns);
    return retval;
}


static void __inline__ 
convert_NxpDefinition_to_Numexp_Definition(NxpDefinition *ndef, Numexp_Definition *cdef)
{
    g_assert(ndef && ndef->name);
    switch (ndef->type)
    {
    case NXP_DEF_VARIABLE:
	cdef->type = Numexp_DEFINITION_VARIABLE;
	cdef->name = CORBA_string_dup(ndef->name);
	break;
    case NXP_DEF_BI_FUNC:
	cdef->type = Numexp_DEFINITION_BUILT_IN_FUNCTION;
	cdef->name = CORBA_string_dup(ndef->name);
	break;
    case NXP_DEF_USER_FUNC:
	cdef->type = Numexp_DEFINITION_USER_FUNCTION;
	cdef->name = CORBA_string_dup(ndef->name);
	break;
    default:
	g_warning("Invalid definition type");
    }
}


static void namespaceListing_add_func(gpointer key, gpointer value, 
				      gpointer user_data) 
{
    Numexp_DefinitionList *r;
    r = (Numexp_DefinitionList*)user_data;
    g_assert(r->_length < r->_maximum);
    convert_NxpDefinition_to_Numexp_Definition((NxpDefinition*) value, 
					       r->_buffer + r->_length++);
}


static Numexp_DefinitionList*
impl_Numexp_Kernel_namespaceListing(PortableServer_Servant servant,
				    const CORBA_char *nsname,
				    CORBA_Environment *ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    const char *def_name = NULL;
    NxpElement *var_elem;
    NxpNameSpace *ns = NULL, *ns1;
    Numexp_Definition *cdef;
    GError *error = NULL;
    int numdefs;
    Numexp_DefinitionList *retval = Numexp_DefinitionList__alloc();

    var_elem = exp_parse_string(nsname, self->context, &error);
    if (error) {
	gerror_to_exception(error, ev);
	return NULL;
    }
    if (NXP_ELEMENT_GET_TYPE(var_elem) != NXP_ELEMENT_TYPE_VARIABLE)
    {
	Numexp_EvalError *eval_err;
	CORBA_string str;
	const char *message = "Numexp::Kernel::registerNamespaceListener:"
	    " bad namespace identifier";

	eval_err = Numexp_EvalError__alloc();
	str = CORBA_string_alloc(strlen(message));
	strcpy(str, message);
	eval_err->detail = str;
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_EvalError,
			    eval_err);
	NXP_ELEMENT_UNREF(var_elem);
	return NULL;
    }
    nxp_identifier_resolve(NXP_ELEMENT_VARIABLE(var_elem)->ident,
			   self->context, &ns, &def_name);
    NXP_ELEMENT_UNREF(var_elem);
    if (!ns || def_name) {
	Numexp_NameError *err;
	err = Numexp_NameError__alloc();
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_NameError, err);
	return NULL;
    }

    nxp_name_space_lock(ns);

    /* Initialize buffer size */
    numdefs = g_hash_table_size(ns->definitions);
    for (ns1 = ns->child; ns1 && ns1->name; ns1 = ns1->next) ++numdefs;
    retval->_length = 0;
    retval->_maximum = numdefs;
    retval->_buffer = Numexp_DefinitionList_allocbuf(numdefs);
    CORBA_sequence_set_release(retval, TRUE);

    /* fill sequence with regular definitions */
    g_hash_table_foreach(ns->definitions, namespaceListing_add_func, retval);

    /* fill in child namespaces as definitions */
    for (ns1 = ns->child; ns1 && ns1->name; ns1 = ns1->next) {
	g_assert(retval->_length < retval->_maximum);
	cdef = retval->_buffer + retval->_length++;
	cdef->type = Numexp_DEFINITION_NAMESPACE;
	cdef->name = CORBA_string_dup(ns1->name);
    }

    retval->_maximum = 0;	/* don't ask */
    nxp_name_space_unlock(ns);
    return retval;
}


static Numexp_Client_ArgList* server_to_client_arglist(GSList *args)
{
    Numexp_Client_ArgList *arglist;
    Numexp_Element *argv;
    int arity;

    arglist = Numexp_Client_ArgList__alloc();
    arity = g_slist_length(args);
    arglist->_length = arity;
    arglist->_buffer = argv = Numexp_Client_ArgList_allocbuf(arity);
    for (; args; args = args->next, ++argv)
	*argv = element2ref(NXP_ELEMENT(args->data));
    CORBA_sequence_set_release(arglist, CORBA_TRUE);
    return arglist;
}


static void listener_func(NxpNameSpace *ns, NxpEvalContext *context, gpointer data)
{
    GString *str;
    NumexpKernel *self = NUMEXP_KERNEL(data);
    CORBA_Environment ev1;
    Numexp_Namespace namespace;

    g_message("data = %p", data);
    str = nxp_name_space_get_full_path_name(ns);
    g_message("Namespace %s has been modified.", str->str);
    namespace.name = CORBA_string_dup(str->str);
    g_string_free(str, TRUE);
    CORBA_exception_init(&ev1);
    Numexp_Client_namespaceNotify(self->client, &namespace, &ev1);
    CORBA_exception_free(&ev1);
}

/* TODO: Define the function and call it under the cd[] definition */
void Numexp_Kernel_to_Client_cwd_change(NxpEvalContext *context)
{
    CORBA_Environment ev1;
    NumexpKernel *self = NUMEXP_KERNEL(nxp_eval_context_get_user_data(context, "kernel"));
    GString *str = nxp_name_space_get_full_path_name(self->context->cwd);
    Numexp_Namespace namespace;

    namespace.name = CORBA_string_dup(str->str);
    g_string_free(str,TRUE);
    CORBA_exception_init(&ev1);
    Numexp_Client_currentNamespaceNotify(self->client, &namespace, &ev1);
    CORBA_exception_free(&ev1);
}


static void impl_Numexp_Kernel_registerNamespaceListener(PortableServer_Servant servant,
							 const CORBA_char * nsname,
							 CORBA_Environment * ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    const char *def_name = NULL;
    NxpElement *var_elem;
    NxpNameSpace *ns = NULL;
    GError *error = NULL;

    var_elem = exp_parse_string(nsname, self->context, &error);
    if (error) {
	gerror_to_exception(error, ev);
	return;
    }
    if (NXP_ELEMENT_GET_TYPE(var_elem) != NXP_ELEMENT_TYPE_VARIABLE)
    {
	Numexp_EvalError *eval_err;
	CORBA_string str;
	const char *message = "Numexp::Kernel::registerNamespaceListener:"
	       " bad namespace identifier";

	eval_err = Numexp_EvalError__alloc();
	str = CORBA_string_alloc(strlen(message));
	strcpy(str, message);
	eval_err->detail = str;
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_EvalError,
			    eval_err);
	NXP_ELEMENT_UNREF(var_elem);
	return;
    }
    nxp_identifier_resolve(NXP_ELEMENT_VARIABLE(var_elem)->ident,
			   self->context, &ns, &def_name);
    NXP_ELEMENT_UNREF(var_elem);
    if (!ns || def_name) {
	Numexp_NameError *err;
	err = Numexp_NameError__alloc();
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_NameError, err);
	return;
    }
    nxp_name_space_add_listener(ns, listener_func, self, self->context);
}


static gboolean listeners_timeout_func(gpointer data)
{
    NumexpKernel *self = NUMEXP_KERNEL(data);
    nxp_name_space_wakeup_listeners(self->context);
    return TRUE;
}


static CORBA_any *
impl_Numexp_Kernel_elementAsAny(PortableServer_Servant servant,
				const Numexp_Element elem,
				CORBA_Environment * ev)
{
    CORBA_any *retval;
    
    retval = CORBA_any__alloc();
    element_to_corba_any(ref2element(elem), retval, ev);
    
    return retval;
}


static Numexp_DefinitionInfo *
impl_Numexp_Kernel_getDefinitionInfo(PortableServer_Servant servant,
				     const CORBA_char * defpath,
				     CORBA_Environment * ev)
{
    Numexp_DefinitionInfo *retval = NULL;
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    const char *def_name = NULL;
    NxpElement *var_elem;
    NxpNameSpace *ns = NULL;
    GError *error = NULL;
    GString *fullpath = NULL;
    NxpDefinition *def = NULL;
    Numexp_EvalError *eval_err;
    CORBA_string str;

    var_elem = exp_parse_string(defpath, self->context, &error);
    if (error) {
	gerror_to_exception(error, ev);
	return NULL;
    }
    if (NXP_ELEMENT_GET_TYPE(var_elem) != NXP_ELEMENT_TYPE_VARIABLE)
    {
	const char *message = "Numexp::Kernel::getDefinitionInfo:"
	    " bad definition identifier";
	
	g_message("%s: %s", __FUNCTION__, message);
	eval_err = Numexp_EvalError__alloc();
	str = CORBA_string_alloc(strlen(message));
	strcpy(str, message);
	eval_err->detail = str;
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_EvalError,
			    eval_err);
	NXP_ELEMENT_UNREF(var_elem);
	return NULL;
    }
    nxp_identifier_resolve(NXP_ELEMENT_VARIABLE(var_elem)->ident,
			   self->context, &ns, &def_name);
    if (ns) {
	fullpath = nxp_name_space_get_full_path_name(ns);
	if (def_name == NULL) {
	    char *help = NULL;
	    help = nxp_help(nxp_identifier_to_string(
                                NXP_ELEMENT_VARIABLE(var_elem)->ident),
                            NULL);
	    retval = Numexp_DefinitionInfo__alloc();
	    retval->description = CORBA_string_dup(help);
	    g_free(help);
	    retval->data._d = Numexp_DEFINITION_NAMESPACE;
	    retval->data._u.value = CORBA_string_dup("");
	} else {
	    def = g_hash_table_lookup(ns->definitions, def_name);
	    if (def) {
		char *help;
		fullpath = g_string_append(fullpath, def_name);
		retval = Numexp_DefinitionInfo__alloc();
		if (def->module)
		    help = nxp_help(fullpath->str, NULL);
		else help = NULL;
		retval->description = CORBA_string_dup(help? help : "No help available.");
		g_free(help);
		retval->data._d = Numexp_DEFINITION_VARIABLE;
		retval->data._u.value = CORBA_string_dup("");
	    } else {
		const char *message = "Numexp::Kernel::getDefinitionInfo:"
		    " no help found";
		g_message("%s: %s", __FUNCTION__, message);
		eval_err = Numexp_EvalError__alloc();
		str = CORBA_string_alloc(strlen(message));
		strcpy(str, message);
		eval_err->detail = str;
		CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_EvalError,
				    eval_err);
		NXP_ELEMENT_UNREF(var_elem);
	    }
	}
	g_string_free(fullpath, TRUE);
    } else {
	const char *message = "Numexp::Kernel::getDefinitionInfo:"
	    " no definition found";
	g_message("%s: %s", __FUNCTION__, message);
	eval_err = Numexp_EvalError__alloc();
	str = CORBA_string_alloc(strlen(message));
	strcpy(str, message);
	eval_err->detail = str;
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_EvalError,
			    eval_err);
	NXP_ELEMENT_UNREF(var_elem);
	return NULL;
    }
   
    return retval;
}


static void
impl_Numexp_Kernel_defineVariable(PortableServer_Servant servant,
				  const CORBA_char * identifier,
				  const Numexp_Element value,
				  CORBA_Environment * ev)
{
    NumexpKernel  *self = NUMEXP_KERNEL(bonobo_object(servant));
    GError        *error = NULL;
    NxpElement    *el;
    NxpIdentifier *id;
    NxpElement    *var_elem;

    var_elem = exp_parse_string(identifier, self->context, &error);
    if (error) {
	gerror_to_exception(error, ev);
	return;
    }
    if (NXP_ELEMENT_GET_TYPE(var_elem) != NXP_ELEMENT_TYPE_VARIABLE)
    {
	Numexp_EvalError *eval_err;
	CORBA_string str;
	const char *message = "Numexp::Kernel::defineVariable:"
	       " bad variable name";

	eval_err = Numexp_EvalError__alloc();
	str = CORBA_string_alloc(strlen(message));
	strcpy(str, message);
	eval_err->detail = str;
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_EvalError,
			    eval_err);
	return;
    }
    id = nxp_identifier_ref(NXP_ELEMENT_VARIABLE(var_elem)->ident);
    NXP_ELEMENT_UNREF(var_elem);
    el = ref2element(value);
    if (!el) {
        g_warning("ref2element(%u) failed!", value);
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_BadParam, NULL);
        return;
    }
    nxp_user_variable_define(id, el, self->context, &error);
    gerror_to_exception(error, ev);
    nxp_identifier_unref(id);
}


static Numexp_Kernel_ModuleList *
impl_Numexp_Kernel_listModules(PortableServer_Servant servant,
			       CORBA_Environment * ev)
{
    /* NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant)); */
    Numexp_Kernel_ModuleList *retval;
    NxpModuleEntry **entries, *entry;
    Numexp_Kernel_Module *module;
    int i;

    entries = nxp_modules_listing();
    g_return_val_if_fail(entries, NULL);

    retval = Numexp_Kernel_ModuleList__alloc();
    retval->_length = 0;
    for (i = 0; entries[i]; ++i) ++retval->_length;    
    retval->_buffer = Numexp_Kernel_ModuleList_allocbuf(retval->_length);
    CORBA_sequence_set_release(retval, CORBA_TRUE);

    for (i = 0; (entry = entries[i]); ++i) {
	module = &retval->_buffer[i];
	module->name        = CORBA_string_dup(entry->name);
	module->location    = CORBA_string_dup(entry->location);
	module->description = CORBA_string_dup(entry->description?:"");
	module->loaded      = entry->loaded? CORBA_TRUE : CORBA_FALSE;
	module->longer_desc = CORBA_string_dup(entry->longer_desc?:"");
    }
    nxp_modules_list_free(entries);
    return retval;
}


static void
impl_Numexp_Kernel_loadModule(PortableServer_Servant servant,
			      const CORBA_char * name, CORBA_Environment * ev)
{
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    nxp_module_load(name, self->context);
}



static Numexp_Element
impl_Numexp_Kernel_elementVoid(PortableServer_Servant servant,
			       CORBA_Environment * ev)
{
   Numexp_Element retval;
   retval = element2ref(nxp_element_void_new());
   return retval;
}


static Numexp_Element
impl_Numexp_Kernel_elementInt(PortableServer_Servant servant,
			      const Numexp_nxp_int_t val,
			      CORBA_Environment * ev)
{
   Numexp_Element retval;
   retval = element2ref(nxp_element_int_new(val));
   return retval;
}


static Numexp_Element
impl_Numexp_Kernel_elementReal(PortableServer_Servant servant,
			       const Numexp_nxp_real_t val,
			       CORBA_Environment * ev)
{
   Numexp_Element retval;
   retval = element2ref(nxp_element_real_new(val));
   return retval;
}


static Numexp_Element
impl_Numexp_Kernel_elementString(PortableServer_Servant servant,
				 const CORBA_char * val,
				 CORBA_Environment * ev)
{
   Numexp_Element retval;
   retval = element2ref(nxp_element_string_new(val));
   return retval;
}

static Numexp_Element
impl_Numexp_Kernel_elementFunction(PortableServer_Servant servant,
				   const CORBA_char * name,
				   const Numexp_Kernel_ArgumentList *arguments,
				   CORBA_Environment * ev)
{
    NumexpKernel         *self = NUMEXP_KERNEL(bonobo_object(servant));
    NxpElement           *elem;
    int                   i;
    NxpElement		*list;
    GSList               *args_list = NULL;
    NxpIdentifier        *ident;
    GError               *error = NULL;

    /* parse identifier */
    ident = nxp_identifier_from_string(name, self->context, &error);
    if (error) {
	gerror_to_exception(error, ev);
	return 0;
    }
 
    /* create argument list */
    for (i = 0; i < arguments->_length; ++i)
	args_list = g_slist_append(args_list, ref2element(arguments->_buffer[i]));

    list = nxp_element_list_new_from_list(args_list);
    elem = nxp_element_function_new(ident, list, NULL);
    NXP_ELEMENT_UNREF(list);
    return element2ref(elem);
}


static CORBA_boolean
impl_Numexp_Kernel_elementIsVoid(PortableServer_Servant servant,
				 const Numexp_Element elem,
				 CORBA_Environment * ev)
{
   CORBA_boolean retval;
   NxpElement *el = ref2element(elem);
   if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_VOID)
       retval = CORBA_TRUE;
   else
       retval = CORBA_FALSE;
   return retval;
}


static CORBA_boolean
impl_Numexp_Kernel_elementIsInt(PortableServer_Servant servant,
				const Numexp_Element elem,
				CORBA_Environment * ev)
{
   CORBA_boolean retval;
   NxpElement *el = ref2element(elem);
   if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_INT)
       retval = CORBA_TRUE;
   else
       retval = CORBA_FALSE;
   return retval;
}

static CORBA_boolean
impl_Numexp_Kernel_elementIsReal(PortableServer_Servant servant,
				 const Numexp_Element elem,
				 CORBA_Environment * ev)
{
   CORBA_boolean retval;
   NxpElement *el = ref2element(elem);
   if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_REAL)
       retval = CORBA_TRUE;
   else
       retval = CORBA_FALSE;
   return retval;
}

static CORBA_boolean
impl_Numexp_Kernel_elementIsString(PortableServer_Servant servant,
				   const Numexp_Element elem,
				   CORBA_Environment * ev)
{
   CORBA_boolean retval;
   NxpElement *el = ref2element(elem);
   if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_STRING)
       retval = CORBA_TRUE;
   else
       retval = CORBA_FALSE;
   return retval;
}

static CORBA_boolean
impl_Numexp_Kernel_elementIsVariable(PortableServer_Servant servant,
				     const Numexp_Element elem,
				     CORBA_Environment * ev)
{
   CORBA_boolean retval;
   NxpElement *el = ref2element(elem);
   if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_VARIABLE)
       retval = CORBA_TRUE;
   else
       retval = CORBA_FALSE;
   return retval;
}

static CORBA_boolean
impl_Numexp_Kernel_elementIsFunction(PortableServer_Servant servant,
				     const Numexp_Element elem,
				     CORBA_Environment * ev)
{
   CORBA_boolean retval;
   NxpElement *el = ref2element(elem);
   if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_FUNCTION)
       retval = CORBA_TRUE;
   else
       retval = CORBA_FALSE;
   return retval;
}

static CORBA_string
impl_Numexp_Kernel_elementVariableGetIdentifier(PortableServer_Servant servant,
						const Numexp_Element elem,
						CORBA_Environment * ev)
{
   CORBA_string retval;
   char *str;
   NxpElement *el = ref2element(elem);
   NxpElementVariable *var;

   if (NXP_ELEMENT_GET_TYPE(el) == NXP_ELEMENT_TYPE_VARIABLE)
   {
       var = NXP_ELEMENT_VARIABLE(el);
       str = nxp_identifier_to_string(var->ident);
       retval = CORBA_string_dup(str);
       g_free(str);
   } else
       retval = CORBA_string_dup("#<!!!ERROR!!! element is not a variable>");
   return retval;
}


static Numexp_Kernel_TopicList*
impl_Numexp_Kernel_getTopicList(PortableServer_Servant servant,
				CORBA_Environment *ev)
{
    GList *list,*l2;
    int i = 0;
    NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    Numexp_Kernel_TopicList *retval;

    list = nxp_help_list(self->context);
    retval = Numexp_Kernel_TopicList__alloc();

    retval->_length = g_list_length(list);
    retval->_buffer = Numexp_Kernel_TopicList_allocbuf(retval->_length);
    l2 = list;
    while (l2) {
	retval->_buffer[i++] = CORBA_string_dup(l2->data);
	g_free(l2->data);
	l2 = g_list_next(l2);
    }
    g_list_free(list);
 
    return retval;
}

static CORBA_string
impl_Numexp_Kernel_getTopicHelp(PortableServer_Servant servant,
				const CORBA_char *topic,
				CORBA_Environment  *ev)
{
   NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));

   return CORBA_string_dup(nxp_help(topic, self->context));
}


static void
_list_variables_visitor(NxpElement *elem,
                        gpointer user_data)
{
    if (NXP_ELEMENT_GET_TYPE(elem) == NXP_ELEMENT_TYPE_VARIABLE)
    {
        NxpIdentifier *id;
        gchar *s;

        id = NXP_ELEMENT_VARIABLE_IDENTIFIER(elem);
        if (nxp_identifier_is_plain(id)) {
            s = nxp_identifier_to_string(id);
            ORBit_sequence_append(user_data, &s);
            g_free(s);
        }
    }
}

static Numexp_Kernel_VariableList*
impl_Numexp_Kernel_elementListVariables(PortableServer_Servant servant,
                                        Numexp_Element elem,
                                        CORBA_Environment * ev)
{
   NumexpKernel *self = NUMEXP_KERNEL(bonobo_object(servant));
    Numexp_Kernel_VariableList *retval;
    NxpElement *el;
    el = ref2element(elem);
    if (!el) {
	g_warning("ref2element(%u) failed!", elem);
	CORBA_exception_set(ev, CORBA_USER_EXCEPTION, ex_Numexp_BadParam, NULL);
        return NULL;
    }

    retval = ORBit_sequence_alloc(TC_Numexp_Kernel_VariableList, 0);
    nxp_element_foreach(el, _list_variables_visitor, retval, self->context);

    return retval;
}


static void numexp_kernel_class_init(NumexpKernelClass *class)
{
    GObjectClass *object_class = (GObjectClass *) class;
    POA_Numexp_Kernel__epv *epv = &class->epv;

    /* Setup ticket sources */
    elem_tksrc = nxp_ticket_source_new();
    def_tksrc  = nxp_ticket_source_new();
    /* . */

    parent_class = g_type_class_peek_parent(class);

    object_class->finalize = numexp_kernel_finalize;

    epv->version                   = impl_Numexp_Kernel_version;

    epv->parseString		   = impl_Numexp_Kernel_parseString;
    epv->elementUnref		   = impl_Numexp_Kernel_elementUnref;
    epv->elementRef		   = impl_Numexp_Kernel_elementRef;
    epv->elementEval		   = impl_Numexp_Kernel_elementEval;
    epv->elementSimplify	   = impl_Numexp_Kernel_elementSimplify;
    epv->elementStr		   = impl_Numexp_Kernel_elementStr;
    epv->elementToMathML	   = impl_Numexp_Kernel_elementToMathML;
    epv->functions		   = impl_Numexp_Kernel_functions;
    epv->setClient	  	   = impl_Numexp_Kernel_setClient;
    epv->registerClientFunction    = impl_Numexp_Kernel_registerClientFunction;
    epv->registerNamespaceListener = impl_Numexp_Kernel_registerNamespaceListener;
    epv->namespaceListing          = impl_Numexp_Kernel_namespaceListing;
    epv->elementAsAny	           = impl_Numexp_Kernel_elementAsAny;
    epv->getDefinitionInfo         = impl_Numexp_Kernel_getDefinitionInfo;
    epv->defineVariable            = impl_Numexp_Kernel_defineVariable;
    epv->listModules               = impl_Numexp_Kernel_listModules;
    epv->loadModule                = impl_Numexp_Kernel_loadModule;

    epv->elementVoid     = impl_Numexp_Kernel_elementVoid;
    epv->elementInt      = impl_Numexp_Kernel_elementInt;
    epv->elementReal     = impl_Numexp_Kernel_elementReal;
    epv->elementString   = impl_Numexp_Kernel_elementString;
    epv->elementFunction = impl_Numexp_Kernel_elementFunction;

    epv->elementIsVoid     = impl_Numexp_Kernel_elementIsVoid;
    epv->elementIsInt      = impl_Numexp_Kernel_elementIsInt;
    epv->elementIsReal     = impl_Numexp_Kernel_elementIsReal;
    epv->elementIsString   = impl_Numexp_Kernel_elementIsString;
    epv->elementIsVariable = impl_Numexp_Kernel_elementIsVariable;
    epv->elementIsFunction = impl_Numexp_Kernel_elementIsFunction;

    epv->elementVariableGetIdentifier = impl_Numexp_Kernel_elementVariableGetIdentifier;
    epv->elementListVariables = impl_Numexp_Kernel_elementListVariables;

    epv->getTopicHelp = impl_Numexp_Kernel_getTopicHelp;
    epv->getTopicList = impl_Numexp_Kernel_getTopicList;

}


