/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include "module.h"
#include "nxp_kernel.h"
#include "eval_context.h"
#include "namespace.h"

/*
   ----------------------
       NxpEvalContext
   ----------------------
*/


/* The eval context owns only anonymous name spaces. Other namespaces
   are just references. */


static gboolean      eval_contexts_initialized = FALSE;
static NxpNameSpace *global_name_space         = NULL;


void nxp_eval_contexts_init(void)
{
    if (eval_contexts_initialized) return;
    eval_contexts_initialized = TRUE;

    global_name_space = nxp_name_space_new(NULL, "(global_name_space)");
}



/**
 * nxp_eval_context_new: -create execution context
 * @root_namespace: the namespace that has the role of global
 * namespace during nxp_element_eval()
 * 
 * Creates an execution context suitable to be passed to nxp_element_eval().
 * 
 * Return value: the new #NxpEvalContext
 **/
NxpEvalContext* nxp_eval_context_new(const char *rcfile)
{
    NxpEvalContext *self;


    nxp_eval_contexts_init();

    self = g_new0(NxpEvalContext, 1);

    self->root_name_space = global_name_space;
    self->cwd = self->root_name_space;
    self->first_anonymous_name_space = NULL;
    self->data = g_hash_table_new_full(g_str_hash, g_str_equal,
				       (GDestroyNotify) g_free,
				       NULL);

    /* TODO: Gerror */
    nxp_types_function_init(self, NULL);
    nxp_module_init(self, NULL);
    nxp_module_load((guchar *)"builtin", self);
    nxp_kernel_load_rc(rcfile, self);

    return self;
}




/**
 * nxp_eval_context_destroy:
 * @self: the context to destroy 
 * 
 * Destroys (and frees) context @self and all anonymous namespaces
 * that belong to it.
 **/
void nxp_eval_context_destroy(NxpEvalContext *self)
{
    g_return_if_fail(self);

    self->root_name_space = NULL;

    if (self->first_anonymous_name_space)
	nxp_name_space_destroy(self->first_anonymous_name_space);

    g_hash_table_destroy(self->data);

    g_free(self);
}


/**
 * nxp_eval_context_push_name_space:
 * @self: a #NxpEvalContext
 * 
 * Create an anonymous name space branching from the context's current
 * global namespace (self->cwd). The new anonymous namespace becomes
 * the new global namespace of the context, and it is owned by the
 * context.
 **/
void nxp_eval_context_push_name_space(NxpEvalContext *self)
{
    NxpNameSpace *ns;
    g_return_if_fail(self);

    ns = nxp_name_space_new(self->cwd, NULL);
    if (!self->first_anonymous_name_space)
	self->first_anonymous_name_space = ns;
    self->cwd = ns;
}



/**
 * nxp_eval_context_pop_name_space:
 * @self: a #NxpEvalContext
 * 
 * Destroy the currently active anonymous namespace. Its parent
 * namespace becomes the new global namespace (the new cwd). This
 * function should only be called to destroy a namespace created with
 * nxp_eval_context_push_name_space().
 **/
void nxp_eval_context_pop_name_space(NxpEvalContext *self)
{
    NxpNameSpace *ns;
    gboolean last_anonymous;

    g_return_if_fail(self);
    last_anonymous = (self->cwd == self->first_anonymous_name_space);
    ns = self->cwd->parent;
    nxp_name_space_destroy(self->cwd);
    if (last_anonymous)
	self->first_anonymous_name_space = NULL;
    self->cwd = ns;
}



void nxp_eval_context_set_user_data(NxpEvalContext *self,
				    const char *dataname,
				    gpointer user_data)
{
    gpointer key, data;
    if (g_hash_table_lookup_extended(self->data, dataname, &key, &data)) {
	g_free(key);
	g_free(data);
    }
    g_hash_table_insert(self->data, g_strdup(dataname), user_data);
}

gpointer nxp_eval_context_get_user_data(NxpEvalContext *self,
					const char *dataname) 
{
    return g_hash_table_lookup(self->data, dataname);
}

