/* -*- Mode: C; c-file-style: "stroustrup"; coding: latin-1-unix -*- */
#include <stdio.h>
#include "nxp_complex.h"
#include "types.h"
#include "types_misc.h"
#include "parser/plugins.h"
#include "errors.h"
#include "ourlibc.h"
#include "element_output.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

static NxpElement* nxp_parse_complex(const gchar *string) 
{
    NxpElement *real = nxp_element_int_new(0);
    NxpElement *img = nxp_element_int_new(1);
    NxpElement *retval = nxp_element_complex_new(real, img);
    NXP_ELEMENT_UNREF(real);
    NXP_ELEMENT_UNREF(img);

    return retval;
}

NxpElement *
nxp_element_complex_new(NxpElement *real, NxpElement *img)
{
    NxpElementComplex *elem;
    g_return_val_if_fail(real == NULL || !NXP_ELEMENT_IS_COMPLEX(real), NULL);
    g_return_val_if_fail(img == NULL || !NXP_ELEMENT_IS_COMPLEX(img), NULL);

    elem = (NxpElementComplex *)nxp_element_alloc(typeinfo, sizeof(NxpElementComplex));
    elem->vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(2, NULL));
    if (real) {
	_element_vector_set(elem->vector, 0, NXP_ELEMENT(real));
	_element_vector_set(elem->vector, 1, NXP_ELEMENT(img));
    }
    return (NxpElement *)elem;
}

NxpElement*
nxp_element_complex_new_from_vector(NxpElement *vector)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(vector), NULL);

    NxpElementComplex *elem;

    if (NXP_ELEMENT_VECTOR(vector)->size == 2) {
	elem = (NxpElementComplex *)nxp_element_alloc(typeinfo, sizeof(NxpElementComplex));
	elem->vector = NXP_ELEMENT_VECTOR(NXP_ELEMENT_REF(vector));
	return NXP_ELEMENT(elem);
    }
    else
	g_warning("Bad size vector's to create a complex...\n");
    return NULL;
}

NxpElement *
nxp_element_complex_new_from_dbl(nxp_complex_t dbl)
{
    NxpElementComplex *elem;

    elem = (NxpElementComplex *)nxp_element_alloc(typeinfo, sizeof(NxpElementComplex));
    elem->vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(2, NULL));
    NXP_COMPLEX_REAL(elem) = nxp_element_real_new(CXREAL(dbl));
    NXP_COMPLEX_IMG(elem) = nxp_element_real_new(CXIMAG(dbl));
    return (NxpElement *)elem;
}

NxpElement *
nxp_element_complex_reduce(NxpElement *e,
			   NxpContext *context, GError **error)
{
    NxpElementComplex *pol = (NxpElementComplex *)e;
    NxpElement *el;
    int i;

    i = 1;
    while (i >= 0 && !nxp_element_is_true(_element_vector_index(pol->vector, i),
		context, error))
	i--;
    if (i == 1)
	el = e;
    else if (i < 0) {
	NXP_ELEMENT_UNREF(e);
	el = nxp_element_int_new(0);
    }
    else if (i == 0) {
	el = _element_vector_get(pol->vector, 0);
	NXP_ELEMENT_UNREF(e);
    }
    else {
	el = NULL;
	g_warning("Very strange error in %s, line %d\n", __FILE__, __LINE__);
    }
    return el;
}

static void nxp_destroy_complex(NxpElement *elem)
{
    NxpElementComplex *pol = NXP_ELEMENT_COMPLEX(elem);

    NXP_ELEMENT_XUNREF(pol->vector);
}

static GString* repr_func(NxpElement *elem, int *precedence)
{
    GString *string;
    GString *old;
    int i, length = 0, diff;
    NxpElementComplex* p;
    NxpElementVector *vector;
    NxpElement *coeff;

    if (precedence) *precedence = 0;
    p = NXP_ELEMENT_COMPLEX(elem);
    vector = p->vector;
    string = g_string_new("");
    for (i = 0; i <= 1; i++) {
	coeff = _element_vector_index(vector, i);
	diff = nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(coeff), 0);
	if (diff) {
	    old = nxp_element_str(coeff); 
	    length += (1 << i);
	    if (diff < 0) {
		if (*string->str)
		    string = g_string_append(string, " - ");
		else
		    string = g_string_append(string, "-");
		if (i == 0 || nxp_element_number_compare_int(
					NXP_ELEMENT_NUMBER(coeff), -1)) {
		    g_string_append_printf(string, "%s", old->str + 1);
		}
	    }
	    else {
		if (*string->str)
		    string = g_string_append(string, " + ");
		if (i == 0 || nxp_element_number_compare_int(
					NXP_ELEMENT_NUMBER(coeff), 1)) {
		    g_string_append_printf(string, "%s", old->str);
		}
	    }
	    if (i == 1)
		string = g_string_append(string, "j");
	    g_string_free(old, TRUE);
	}
    }
    if (*string->str == 0)
	string = g_string_append(string, "0");
    if (precedence) {
	switch (length) {
	    case 0:
	    case 1:
		* precedence = 0;
		break;
	    case 2:
		* precedence = 40;
		break;
	    case 3:
		* precedence = 50;
	}
    }
    return string;
}

static NxpElement* mul(NxpOp op, NxpElement* a, NxpElement* b,
		       NxpEvalContext *context, GError **error)
{
    NxpElement *pol, *prod1, *prod2;
    NxpElementVector *va, *vb;

    va = NXP_ELEMENT_COMPLEX(a)->vector;
    vb = NXP_ELEMENT_COMPLEX(b)->vector;
    pol = nxp_element_complex_new(NULL, NULL);

    prod1 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 _element_vector_index(va, 0),
					 _element_vector_index(vb, 0),
					 context, error);
    prod2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 _element_vector_index(va, 1),
					 _element_vector_index(vb, 1),
					 context, error);
    NXP_COMPLEX_REAL(pol)
		= nxp_element_tree_simplify_op(NXP_OP_MINUS, prod1, prod2,
					       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    prod1 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 _element_vector_index(va, 0),
					 _element_vector_index(vb, 1),
					 context, error);
    prod2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 _element_vector_index(va, 1),
					 _element_vector_index(vb, 0),
					 context, error);
    NXP_COMPLEX_IMG(pol)
		= nxp_element_tree_simplify_op(NXP_OP_PLUS, prod1, prod2,
					       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    return nxp_element_complex_reduce(pol, context, error);
}

static NxpElement* mul_eval(NxpOp op, NxpElement* a, NxpElement* b,
			    NxpEvalContext *context, GError **error)
{
    NxpElement *pol, *prod1, *prod2;
    NxpElementVector *va, *vb;

    va = NXP_ELEMENT_COMPLEX(a)->vector;
    vb = NXP_ELEMENT_COMPLEX(b)->vector;
    pol = nxp_element_complex_new(NULL, NULL);

    prod1 = nxp_element_tree_eval_op(NXP_OP_STAR,
					 _element_vector_index(va, 0),
					 _element_vector_index(vb, 0),
					 context, error);
    prod2 = nxp_element_tree_eval_op(NXP_OP_STAR,
					 _element_vector_index(va, 1),
					 _element_vector_index(vb, 1),
					 context, error);
    NXP_COMPLEX_REAL(pol)
		= nxp_element_tree_eval_op(NXP_OP_MINUS, prod1, prod2,
					       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    prod1 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     _element_vector_index(va, 0),
				     _element_vector_index(vb, 1),
				     context, error);
    prod2 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     _element_vector_index(va, 1),
				     _element_vector_index(vb, 0),
				     context, error);
    NXP_COMPLEX_IMG(pol)
		= nxp_element_tree_eval_op(NXP_OP_PLUS, prod1, prod2,
					   context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    return nxp_element_complex_reduce(pol, context, error);
}

static NxpElement* div(NxpOp op, NxpElement* a, NxpElement* b,
		       NxpEvalContext *context, GError **error)
{
    NxpElement *vec, *vec1, *prod1, *prod2;
    NxpElement *den, *retval;

    vec = nxp_element_vector_new(2, NULL);

    prod1 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 NXP_COMPLEX_REAL(a),
					 NXP_COMPLEX_REAL(b),
					 context, error);
    prod2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 NXP_COMPLEX_IMG(a),
					 NXP_COMPLEX_IMG(b),
					 context, error);
    _element_vector_index(NXP_ELEMENT_VECTOR(vec), 0)
		= nxp_element_tree_simplify_op(NXP_OP_PLUS, prod1, prod2,
					       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    prod1 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 NXP_COMPLEX_REAL(b),
					 NXP_COMPLEX_IMG(a),
					 context, error);
    prod2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 NXP_COMPLEX_REAL(a),
					 NXP_COMPLEX_IMG(b),
					 context, error);
    _element_vector_index(NXP_ELEMENT_VECTOR(vec), 1)
		= nxp_element_tree_simplify_op(NXP_OP_MINUS, prod1, prod2,
					       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    prod1 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 NXP_COMPLEX_REAL(b),
					 NXP_COMPLEX_REAL(b),
					 context, error);
    prod2 = nxp_element_tree_simplify_op(NXP_OP_STAR,
					 NXP_COMPLEX_IMG(b),
					 NXP_COMPLEX_IMG(b),
					 context, error);
    den = nxp_element_tree_simplify_op(NXP_OP_PLUS, prod1, prod2,
				       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);
    vec1 = nxp_element_tree_simplify_op(NXP_OP_SLASH, vec, den, context, error);
    retval = nxp_element_complex_new_from_vector(vec1);
    NXP_ELEMENT_UNREF(vec1);
    NXP_ELEMENT_UNREF(den);

    return nxp_element_complex_reduce(retval, context, error);
}

static NxpElement* div_eval(NxpOp op, NxpElement* a, NxpElement* b,
			    NxpEvalContext *context, GError **error)
{
    NxpElement *vec, *vec1, *prod1, *prod2;
    NxpElement *den, *retval;

    vec = nxp_element_vector_new(2, NULL);

    prod1 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     NXP_COMPLEX_REAL(a),
				     NXP_COMPLEX_REAL(b),
				     context, error);
    prod2 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     NXP_COMPLEX_IMG(a),
				     NXP_COMPLEX_IMG(b),
				     context, error);
    _element_vector_index(NXP_ELEMENT_VECTOR(vec), 0)
		= nxp_element_tree_eval_op(NXP_OP_PLUS, prod1, prod2,
					       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    prod1 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     NXP_COMPLEX_REAL(b),
				     NXP_COMPLEX_IMG(a),
				     context, error);
    prod2 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     NXP_COMPLEX_REAL(a),
				     NXP_COMPLEX_IMG(b),
				     context, error);
    _element_vector_index(NXP_ELEMENT_VECTOR(vec), 1)
		= nxp_element_tree_eval_op(NXP_OP_MINUS, prod1, prod2,
					       context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    prod1 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     NXP_COMPLEX_REAL(b),
				     NXP_COMPLEX_REAL(b),
				     context, error);
    prod2 = nxp_element_tree_eval_op(NXP_OP_STAR,
				     NXP_COMPLEX_IMG(b),
				     NXP_COMPLEX_IMG(b),
				     context, error);
    den = nxp_element_tree_eval_op(NXP_OP_PLUS, prod1, prod2,
				   context, error);
    NXP_ELEMENT_UNREF(prod1);
    NXP_ELEMENT_UNREF(prod2);

    vec1 = nxp_element_tree_eval_op(NXP_OP_SLASH, vec, den, context, error);
    retval = nxp_element_complex_new_from_vector(vec1);
    NXP_ELEMENT_UNREF(vec1);
    NXP_ELEMENT_UNREF(den);

    return nxp_element_complex_reduce(retval, context, error);
}

static NxpElement* nxp_element_complex_opposed(NxpElementNumber *self)
{
    NxpElement *retval;

    retval = nxp_element_complex_new(NULL, NULL);

    NXP_COMPLEX_REAL(retval) = nxp_element_number_opposed(
				NXP_ELEMENT_NUMBER(NXP_COMPLEX_REAL(self)));
    NXP_COMPLEX_IMG(retval) = nxp_element_number_opposed(
				NXP_ELEMENT_NUMBER(NXP_COMPLEX_IMG(self)));
    return retval;
}

static NxpElement* minus(NxpOp op, NxpElement* a, NxpElement* b,
			 NxpEvalContext *context, GError **error)
{
    return nxp_element_complex_opposed(NXP_ELEMENT_NUMBER(b));
}

static gboolean truth_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    return (nxp_element_is_true(NXP_COMPLEX_REAL(elem), context, error)
	    || nxp_element_is_true(NXP_COMPLEX_IMG(elem), context, error));
}

static NxpElement* add(NxpOp op, NxpElement* a, NxpElement* b,
		       NxpEvalContext *context, GError **error)
{
    NxpElement *pol;

    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(b), NULL);

    pol = nxp_element_complex_new(NULL, NULL);

    NXP_COMPLEX_REAL(pol) = nxp_element_tree_simplify_op(
				    op,
				    NXP_COMPLEX_REAL(a),
				    NXP_COMPLEX_REAL(b),
				    context, error);
    NXP_COMPLEX_IMG(pol) = nxp_element_tree_simplify_op(
				    op,
				    NXP_COMPLEX_IMG(a),
				    NXP_COMPLEX_IMG(b),
				    context, error);
    return nxp_element_complex_reduce(pol, context, error);
}

static NxpElement* eq(NxpOp op, NxpElement *a, NxpElement *b,
		      NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(b), NULL);

    if (nxp_element_eq(NXP_ELEMENT(NXP_ELEMENT_COMPLEX(a)->vector),
		       NXP_ELEMENT(NXP_ELEMENT_COMPLEX(b)->vector),
		       context, error))
	return nxp_element_int_new(1);
    else
	return nxp_element_int_new(0);
}

static int nxp_element_complex_compare_int(NxpElementNumber *self, nxp_int_t nb)
{
    int cpx = nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(NXP_COMPLEX_IMG(self)), 0);
    int real = nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(NXP_COMPLEX_REAL(self)), nb);

    /* The idea is to know if the first number written in real + i*cpx
     * is positive or not */
    if (cpx ==0)
	return real;
    else {
	if (real)
	    return real;
	else
	    return cpx;
    }
}

nxp_complex_t nxp_element_complex_get_value(NxpElementComplex *self)
{
    nxp_real_t real, img;

    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(self), 0);

    real = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_REAL(self)));
    img = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_IMG(self)));

    return real + img * 1i;
}

static NxpElement* eval_func(NxpElement *self, NxpEvalContext *context, GError **error)
{
    NxpElement *real, *img;
    NxpElement *retval;

    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(self), 0);

    real = nxp_element_eval(NXP_COMPLEX_REAL(self), context, error);
    img = nxp_element_eval(NXP_COMPLEX_IMG(self), context, error);

    retval = nxp_element_complex_new(real, img);
    NXP_ELEMENT_UNREF(real);
    NXP_ELEMENT_UNREF(img);
    return retval;
}

#define BIND(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_COMPLEX, \
							 NXP_ELEMENT_TYPE_COMPLEX, \
    			                                 eval, symp, 0, NULL);
#define BIND_LEFT(eval,symp,op)    nxp_bind_operator_function(op, 0,         \
                                                         NXP_ELEMENT_TYPE_COMPLEX, \
   			                                 eval, symp, 0, NULL);

NxpElementType nxp_element_complex_get_type(void)
{
    static int nxp_element_type_complex = 0;

    if (nxp_element_type_complex == 0) {
	nxp_element_type_complex = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
				nxp_element_type_complex, /* The new type */
				sizeof(NxpElementNumberTypeInfo), /* The size of our typeinfo */
				NXP_ELEMENT_TYPE_NUMBER, /* the type of the parent */
				sizeof(NxpElementNumberTypeInfo)); /* The size of the parent's typeinfo */
    }
    return nxp_element_type_complex;
}

void nxp_element_complex_type_init() 
{
    NxpElementNumberTypeInfo *nb_typeinfo;
    typeinfo = nxp_element_type_get_info(nxp_element_complex_get_type());
    nb_typeinfo = (NxpElementNumberTypeInfo *)typeinfo;

    typeinfo->name = g_strdup("complex");
    typeinfo->destroy_func = nxp_destroy_complex;
    typeinfo->eval_func = eval_func;
    typeinfo->simplify_func = nxp_eval_return_self;
    typeinfo->str_func = repr_func;
    typeinfo->repr_func = repr_func;
    typeinfo->truth_func = truth_func;

    nb_typeinfo->compare_int = nxp_element_complex_compare_int;
    nb_typeinfo->opposed = nxp_element_complex_opposed;

    nxp_parser_register_func(NXP_ATOM_CLASS_COMPLEX, nxp_parse_complex, 50);

    BIND(mul_eval, mul, NXP_OP_STAR);
    BIND(add, add, NXP_OP_PLUS);
    BIND(add, add, NXP_OP_MINUS);
    BIND_LEFT(minus, minus, NXP_OP_MINUS_LEFT);

    BIND(eq, eq, NXP_OP_EQ2);
    BIND(div_eval, div, NXP_OP_SLASH);
    //    NXP_OP_HAT

    /// not implementable ///////
    //
    //    NXP_OP_DOT3
    //    NXP_OP_PLUS2_LEFT
    //    NXP_OP_MINUS2_LEFT
    //    NXP_OP_PLUS2_RIGHT
    //    NXP_OP_MINUS2_RIGHT
    /////////////////////////////

    // >,<,>= and <= compares SIZE
//    BIND( gt, gt, NXP_OP_GT);
//    BIND( le, le, NXP_OP_LT_EQ);
//    BIND( lt, lt, NXP_OP_LT);
//    BIND( ge, ge, NXP_OP_GT_EQ);

    // == and != compares ALL
//    BIND( ne, NULL, NXP_OP_EXCL_EQ);

    // These two compares sizes, only
//    BIND( and, and, NXP_OP_SLASH_BSLASH);
//    BIND( or,  or,  NXP_OP_BSLASH_SLASH);
//    BIND_LEFT( not, not, NXP_OP_TILDE_LEFT);

    ///// not implementable ///
    //
    //    NXP_OP_EXCL_RIGHT
    //    NXP_OP_EQ
    //    NXP_OP_COLON_EQ
    ///////////////////////////
}

void nxp_element_complex_function_init(NxpEvalContext *context, GError **error)
{
}
