/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include <math.h>

#include "int.h"
#include "real.h"
#include "frac.h"
#include "tree.h"
#include "op_bindings.h"
#include "types_misc.h"
#include "errors.h"
#include "i18n.h"

#define INT_REAL(func, op)    static NxpElement* func(NxpOp op, NxpElement* a, NxpElement *b, \
			    NxpEvalContext *context, GError **error) { \
    mpfr_t val; \
    NxpElement* retval;\
\
    mpfr_init_set_z(val, NXP_ELEMENT_INT(a)->value, GMP_RNDN); \
    mpfr_##op(val, val, NXP_ELEMENT_REAL(b)->value, GMP_RNDN); \
    retval = nxp_element_real_new_from_mpfr(val); \
    return retval; \
}

INT_REAL(add_i_r, add);
INT_REAL(mul_i_r, mul);
INT_REAL(sub_i_r, sub);

static NxpElement *sub_r_i(NxpOp op, NxpElement* a, NxpElement *b,
			   NxpEvalContext *context, GError **error) {
    mpfr_t val;
    NxpElement *retval;

    mpfr_init(val);
    mpfr_sub_z(val, NXP_ELEMENT_REAL(a)->value, NXP_ELEMENT_INT(b)->value, GMP_RNDN);
    retval = nxp_element_real_new_from_mpfr(val);
    return retval;
}

static int
_cmp_r_i(NxpElement* a, NxpElement *b)
{
    mpfr_t z;
    int retval;

    mpfr_init_set_z(z, NXP_ELEMENT_INT(b)->value, GMP_RNDN);
    retval = (mpfr_cmp(NXP_ELEMENT_REAL(a)->value, z));
    mpfr_clear(z);
    return retval;
}

static NxpElement *eq2(NxpOp op, NxpElement* a, NxpElement* b,
		       NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(b, a) == 0);
}

static NxpElement *neq2(NxpOp op, NxpElement* a, NxpElement* b,
		       NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(b, a) != 0);
}

static NxpElement *gt_r_i(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(a, b) > 0);
}

static NxpElement *lt_r_i(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(a, b) < 0);
}

static NxpElement *ge_r_i(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(a, b) >= 0);
}

static NxpElement *le_r_i(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(a, b) <= 0);
}

static NxpElement *gt_i_r(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(b, a) < 0);
}

static NxpElement *lt_i_r(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(b, a) > 0);
}

static NxpElement *ge_i_r(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(b, a) <= 0);
}

static NxpElement *le_i_r(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(_cmp_r_i(b, a) >= 0);
}


static NxpElement *div_i_r(NxpOp op, NxpElement* a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    mpfr_t val;
    NxpElement *retval;
    if (mpfr_cmp_ui(NXP_ELEMENT_REAL(b)->value, 0) != 0) {
	mpfr_init_set_z(val, NXP_ELEMENT_INT(a)->value, GMP_RNDN);
	mpfr_div(val, val, NXP_ELEMENT_REAL(b)->value, GMP_RNDN);
	retval = nxp_element_real_new_from_mpfr(val);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO, _("divide by zero"));
	retval = NULL;
    }
    return retval;
}

static NxpElement *div_r_i(NxpOp op, NxpElement* a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    mpfr_t val;
    NxpElement *retval;
    if (mpz_cmp_ui(NXP_ELEMENT_INT(b)->value, 0) != 0) {
	mpfr_init(val);
	mpfr_div_z(val, NXP_ELEMENT_REAL(a)->value, NXP_ELEMENT_INT(b)->value, GMP_RNDN);
	retval = nxp_element_real_new_from_mpfr(val);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO, _("divide by zero"));
	retval = NULL;
    }
    return retval;
}

static NxpElement* hat_r_i(NxpOp op, NxpElement* a, NxpElement *b,
				NxpEvalContext *context, GError **error)
{
    mpfr_t r;
    nxp_int_t pow;
    NxpElement *el;

    mpfr_init(r);
    if (mpz_fits_slong_p(NXP_ELEMENT_INT(b)->value)) {
	pow = mpz_get_si(NXP_ELEMENT_INT(b)->value);
	mpfr_pow_si(r, NXP_ELEMENT_REAL(a)->value, pow, GMP_RNDN);
    }
    else {
	mpfr_set_z(r, NXP_ELEMENT_INT(b)->value, GMP_RNDN);
	mpfr_pow(r, NXP_ELEMENT_REAL(a)->value, r, GMP_RNDN);
    }
    el = nxp_element_real_new_from_mpfr(r);
    return el;
}

static NxpElement* hat_i_r(NxpOp op, NxpElement* a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    NxpElement *el;
    mpfr_t r;

    mpfr_init_set_z(r, NXP_ELEMENT_INT(a)->value, GMP_RNDN);
    mpfr_pow(r, r, NXP_ELEMENT_REAL(b)->value, GMP_RNDN);
    el = nxp_element_real_new_from_mpfr(r);
    return el;
}

#define BIND_I_R(op,simp) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_INT, NXP_ELEMENT_TYPE_REAL, \
   			                              simp, simp,  0, NULL);
#define BIND_R_I(op,simp) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_REAL, NXP_ELEMENT_TYPE_INT, \
   			                              simp, simp,  0, NULL);

#define BIND_SYM(op,simp) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_INT, \
							     NXP_ELEMENT_TYPE_REAL, \
   			                              simp, simp,  NXP_OP_BIND_SYMMETRIC, NULL);


void nxp_element_real_int_operator_init(void)
{
    BIND_SYM(NXP_OP_PLUS, add_i_r);
    BIND_SYM(NXP_OP_STAR, mul_i_r);

    BIND_I_R(NXP_OP_MINUS, sub_i_r);
    BIND_R_I(NXP_OP_MINUS, sub_r_i);

    BIND_I_R(NXP_OP_SLASH, div_i_r);
    BIND_R_I(NXP_OP_SLASH, div_r_i);

    BIND_R_I(NXP_OP_HAT, hat_r_i);
    BIND_I_R(NXP_OP_HAT, hat_i_r);
//    BIND_F_I(NXP_OP_HAT, simp_f_r_hat);
    BIND_SYM(NXP_OP_EQ2, eq2);
    BIND_SYM(NXP_OP_EXCL_EQ, neq2);
    BIND_R_I(NXP_OP_GT, gt_r_i);
    BIND_R_I(NXP_OP_GT_EQ, ge_r_i);
    BIND_R_I(NXP_OP_LT, lt_r_i);
    BIND_R_I(NXP_OP_LT_EQ, le_r_i);
    BIND_I_R(NXP_OP_GT, gt_i_r);
    BIND_I_R(NXP_OP_GT_EQ, ge_i_r);
    BIND_I_R(NXP_OP_LT, lt_i_r);
    BIND_I_R(NXP_OP_LT_EQ, le_i_r);
/*   BIND_I_R(hat_int_real, NXP_OP_HAT); */

/*   BIND_R_I(and_real_int, NXP_OP_SLASH_BSLASH); */
/*   BIND_I_R(and_int_real, NXP_OP_SLASH_BSLASH); */

/*   BIND_R_I(or_real_int, NXP_OP_BSLASH_SLASH); */
/*   BIND_I_R(or_int_real, NXP_OP_BSLASH_SLASH); */

}
