/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include <string.h>
#include "variable.h"
#include "int.h"
#include "tree.h"
#include "mpoly.h"
#include "identifier.h"
#include "variables.h"
#include "errors.h"
#include "op_bindings.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;


/**
 * nxp_element_variable_new:
 * @ident: the identifier of the variable
 * 
 * Create a new NxpElementVariable
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_variable_new(NxpIdentifier *ident)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementVariable));
    NXP_ELEMENT_VARIABLE(elem)->ident = nxp_identifier_ref(ident);
    return elem;
}




/**
 * nxp_element_variable_new_simple:
 * @name: const string with the name of the variable; no paths allowed, only one component.
 * 
 * Create variable element from c string instead of identifier
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_variable_new_simple(const char *name)
{
    NxpIdentifier *ident;
    NxpElement *el;
    ident = nxp_identifier_new_simple(name);
    el = nxp_element_variable_new(ident);
    nxp_identifier_unref(ident);
    return el;
}

static NxpElement* nxp_variable_derive(NxpElement *a, NxpIdentifier *ident, 
				       NxpEvalContext *context, GError **error)
{
/*    NxpElement *simp, *retval;

    simp = nxp_element_simplify(a, context, error);
    if (NXP_ELEMENT_IS_VARIABLE(simp)) {
	retval = nxp_element_int_new((nxp_identifier_equal(
					    NXP_ELEMENT_VARIABLE(simp)->ident,
					    ident, context)) ? 1 : 0);
    }
    else
	retval = nxp_element_derive(simp, ident, context, error);
    NXP_ELEMENT_UNREF(simp);
    return retval;*/
    NxpElement *content, *retval;

    if (nxp_identifier_equal(NXP_ELEMENT_VARIABLE(a)->ident, ident, context))
	return nxp_element_int_new(1);
    else {
	content = nxp_variable_get(NXP_ELEMENT_VARIABLE(a)->ident, context);
	if (content) {
	    retval = nxp_element_derive(content, ident, context, error);
	    NXP_ELEMENT_UNREF(content);
	    return retval;
	}
    }
    return nxp_element_int_new(0);
}


static NxpElement* eval_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElementVariable *var = NXP_ELEMENT_VARIABLE(elem);
    GError *err = NULL;
    NxpElement *value, *eval;
    value = nxp_variable_get(var->ident, context);
    if (!value) {
	char *name = nxp_identifier_to_string(var->ident);
	g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
		    _("undefined variable '%s'"), name);
	g_free(name);
	return NULL;
    }

    eval = nxp_element_eval(value, context, &err);
    if (err) {
	NXP_ELEMENT_XUNREF(eval);
	NXP_ELEMENT_UNREF(value);
	g_propagate_error(error, err);
	return NULL;
    } else {
	NXP_ELEMENT_XUNREF(value);
	return eval;
    }
}


static NxpElement* simplify_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElementVariable *var = NXP_ELEMENT_VARIABLE(elem);
    NxpElement *value, *simp;
    value = nxp_variable_get(var->ident, context);
    if (value) {
	GError *error1 = NULL;
	simp = nxp_element_simplify(value, context, &error1);
	NXP_ELEMENT_UNREF(value);
	if (error1) g_propagate_error(error, error1);
	return simp;
    } else
	return NXP_ELEMENT_REF(elem);
}


static GString* str_func(NxpElement *elem, int *prec)
{
    if (prec) *prec = 0;
    return nxp_identifier_to_gstring(NXP_ELEMENT_VARIABLE(elem)->ident);
}

static gboolean truth_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElement *content;

    content = nxp_variable_get(NXP_ELEMENT_VARIABLE(elem)->ident, context);
    if (!content) {
	GString *str = str_func(elem, NULL);
	g_set_error(error, nxp_error_domain, NXP_ERROR_UNDECIDABLE, _("Cannot decide if %s is true."), str->str);
	g_string_free(str, TRUE);
	return FALSE;
    }
    else if (nxp_element_is_true(content, context, error)) {
	NXP_ELEMENT_UNREF(content);
	return TRUE;
    }
    NXP_ELEMENT_UNREF(content);
    return FALSE;
}

static gboolean dependson_func(NxpElement *self, NxpElement *variable,
			       NxpEvalContext *context, GError **error)
{
    if (nxp_identifier_equal(NXP_ELEMENT_VARIABLE(self)->ident,
			     NXP_ELEMENT_VARIABLE(variable)->ident, context))
	return TRUE;
    else {
	NxpElement *content = nxp_variable_get(NXP_ELEMENT_VARIABLE(self)->ident, context);
	if (content) {
	    gboolean retval;

	    retval = nxp_element_depends_on(content, variable, context, error);
	    NXP_ELEMENT_UNREF(content);
	    return retval;
	}
	else
	    return FALSE;
    }
}

static NxpElement* eval_assign(NxpOp op, NxpElement *a, NxpElement *b, 
			       NxpEvalContext *context, GError **error)
{
    NxpElement *value;
    GError *error1 = NULL;
    char *varname;

    varname = nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(a)->ident);
    if (varname[0] == '_' || strstr(varname, "::_")) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("You can't assign to a numexp private variable"));
	    return NULL;
    }
    g_free(varname);

    value = nxp_element_eval(b, context, &error1);
    if (error1) {
	NXP_ELEMENT_XUNREF(value);
	g_propagate_error(error, error1);
	return NULL;
    }
    if (NXP_ELEMENT_GET_TYPE(value) != NXP_ELEMENT_TYPE_VARIABLE ||
	    !nxp_identifier_equal(NXP_ELEMENT_VARIABLE(a)->ident,
				  NXP_ELEMENT_VARIABLE(value)->ident, context))
	nxp_user_variable_define(NXP_ELEMENT_VARIABLE(a)->ident, value, context, &error1);
    if (error1) {
	NXP_ELEMENT_XUNREF(value);
	g_propagate_error(error, error1);
	return NULL;
    }
    return value;
}

static NxpElement* simp_assign(NxpOp op, NxpElement *a, NxpElement *b, 
			       NxpEvalContext *context, GError **error)
{
    NxpElement *value;
    GError *error1 = NULL;
    char *varname;

    varname = nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(a)->ident);
    if (varname[0] == '_' || strstr(varname, "::_")) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("You can't assign to a numexp private variable"));
	    return NULL;
    }
    g_free(varname);

    value = nxp_element_simplify(b, context, &error1);
    if (error1) {
	NXP_ELEMENT_XUNREF(value);
	g_propagate_error(error, error1);
	return NULL;
    }
    if (nxp_element_depends_on(value, a, context, error)) {
	
	g_set_error(error, nxp_error_domain, NXP_ERROR_RECURSIVE_ASSIGNMENT, _("recursive assignment"));
	NXP_ELEMENT_UNREF(value);
	return NULL;
    }

    nxp_user_variable_define(NXP_ELEMENT_VARIABLE(a)->ident, value, context, &error1);
    if (error1) {
	g_propagate_error(error, error1);
	NXP_ELEMENT_XUNREF(value);
	return NULL;
    }
    return value;
}


static NxpElement* substitute_func(NxpElement *elem, NxpElement *var,
				   NxpElement *el, NxpEvalContext *context,
				   GError **error)
{
    if (nxp_identifier_equal(NXP_ELEMENT_VARIABLE(elem)->ident,
			     NXP_ELEMENT_VARIABLE(var)->ident, context))
	return NXP_ELEMENT_REF(el);
    else
	return NXP_ELEMENT_REF(elem);
}

static NxpElement* eval_store(NxpOp op, NxpElement *a, NxpElement *b, 
			      NxpEvalContext *context, GError **error)
{
    NxpElement *value;
    GError *erro = NULL;
    char *varname;

    varname = nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(a)->ident);
    if (varname[0] == '_' || strstr(varname, "::_")) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("You can't assign to a numexp private variable"));
	    return NULL;
    }
    g_free(varname);

    value = NXP_ELEMENT_REF(b);
    if (nxp_element_depends_on(value, a, context, error)) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_RECURSIVE_ASSIGNMENT, _("recursive assignment"));
	NXP_ELEMENT_UNREF(value);
	return NULL;
    }

    nxp_user_variable_define(NXP_ELEMENT_VARIABLE(a)->ident, value, context, &erro);
    if (erro) {
	g_propagate_error(error, erro);
	NXP_ELEMENT_XUNREF(value);
	return NULL;
    }
    return value;
}

static NxpElement* minus_left(NxpOp op, NxpElement *a, NxpElement *b,
			      NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    NxpElement *args, *mon, *coeff, *vector;

    args = nxp_element_list_new();
    nxp_element_list_append(args, b);
    coeff = nxp_element_int_new(-1);
    vector = nxp_element_vector_new(1, NULL);
    _element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = nxp_element_int_new(1);
    mon = nxp_monomial_new(coeff, vector);
    retval = nxp_element_mpoly_new_simple(mon, args);
    NXP_ELEMENT_UNREF(coeff);
    NXP_ELEMENT_UNREF(vector);
    NXP_ELEMENT_UNREF(mon);
    NXP_ELEMENT_UNREF(args);
    return retval;
}

static NxpElement* minus_left_eval(NxpOp op, NxpElement *a, NxpElement *b,
				   NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    NxpElement *el = nxp_element_eval(b, context, error);
    if (el) {
        retval = nxp_element_tree_eval_op(op, a, el, context, error);
	NXP_ELEMENT_UNREF(el);
    }
    else
	retval = NULL;
    return retval;
}

static NxpElement* simp_mul(NxpOp op, NxpElement *a, NxpElement *b,
			    NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    NxpElementVector *vector;
    NxpElement *mon, *args;
    NxpElement *one;

    one = nxp_element_int_new(1);
    if (nxp_element_eq(a, b, context, error)) {
	/* One variable polynomial */
	vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(1, NULL));
	_element_vector_index(vector, 0) = nxp_element_int_new(2);

	mon = nxp_monomial_new(one, NXP_ELEMENT(vector));

	args = nxp_element_list_new_simple(a);
    }
    else {
	/* Multivariable polynomial */
	vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(2, NULL));
	_element_vector_set(vector, 0, one);
	_element_vector_set(vector, 1, one);

	mon = nxp_monomial_new(one, NXP_ELEMENT(vector));

	args = nxp_element_list_new_simple(a);
	nxp_element_list_append(args, b);
    }
    NXP_ELEMENT_UNREF(vector);
    NXP_ELEMENT_UNREF(one);
    retval = nxp_element_mpoly_new_simple(mon, args);
    NXP_ELEMENT_UNREF(mon);
    NXP_ELEMENT_UNREF(args);
    return retval;
}

static NxpElement* simp_div(NxpOp op, NxpElement *a, NxpElement *b,
			    NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    NxpElementVector *vector;
    NxpElement *mon, *args;
    NxpElement *one;

    one = nxp_element_int_new(1);
    if (nxp_element_eq(a, b, context, error))
	return one;
    else {
	/* Multivariable polynomial */
	vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(2, NULL));
	_element_vector_set(vector, 0, one);
	_element_vector_index(vector, 1) = nxp_element_int_new(-1);

	mon = nxp_monomial_new(one, NXP_ELEMENT(vector));

	args = nxp_element_list_new_simple(a);
	nxp_element_list_append(args, b);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(one);
	retval = nxp_element_mpoly_new_simple(mon, args);
	NXP_ELEMENT_UNREF(mon);
	NXP_ELEMENT_UNREF(args);
    }
    return retval;
}

static NxpElement* inc_eval(NxpOp op, NxpElement *a, NxpElement *b,
			    NxpEvalContext *context, GError **error)
{
    GError *err = NULL;
    NxpElement *one = nxp_element_int_new(1), *retval;
    NxpElement *el = nxp_element_tree_eval_op(NXP_OP_PLUS, b, one, context, &err);
    NXP_ELEMENT_UNREF(one);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    retval = eval_assign(NXP_OP_EQ, b, el, context, &err);
    NXP_ELEMENT_UNREF(el);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    return retval;
}

static NxpElement* dec_eval(NxpOp op, NxpElement *a, NxpElement *b,
			    NxpEvalContext *context, GError **error)
{
    GError *err = NULL;
    NxpElement *one = nxp_element_int_new(1), *retval;
    NxpElement *el = nxp_element_tree_eval_op(NXP_OP_MINUS, b, one, context, &err);
    NXP_ELEMENT_UNREF(one);

    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    retval = eval_assign(NXP_OP_EQ, b, el, context, &err);
    NXP_ELEMENT_UNREF(el);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    return retval;
}

static NxpElement* inc_simp(NxpOp op, NxpElement *a, NxpElement *b,
			    NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_VARIABLE(b), NULL);
    GError *err = NULL;
    NxpElement *one = nxp_element_int_new(1), *retval;
    NxpElement *el = nxp_element_tree_simplify_op(NXP_OP_PLUS, b, one, context, &err);
    NXP_ELEMENT_UNREF(one);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    retval = simp_assign(NXP_OP_EQ, b, el, context, &err);
    NXP_ELEMENT_UNREF(el);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    return retval;
}

static NxpElement* dec_simp(NxpOp op, NxpElement *a, NxpElement *b,
			    NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_VARIABLE(b), NULL);
    GError *err = NULL;
    NxpElement *one = nxp_element_int_new(1), *retval;
    NxpElement *el = nxp_element_tree_simplify_op(NXP_OP_MINUS, b, one, context, &err);
    NXP_ELEMENT_UNREF(one);

    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    retval = simp_assign(NXP_OP_EQ, b, el, context, &err);
    NXP_ELEMENT_UNREF(el);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    return retval;
}

static NxpElement* simp_add(NxpOp op, NxpElement *a, NxpElement *b,
			    NxpEvalContext *context, GError **error)
{
    NxpElement *mon1, *mon2;
    NxpElementVector *vector;
    NxpElement *el;
    NxpElement *args, *mon;

    if (nxp_identifier_equal(NXP_ELEMENT_VARIABLE(a)->ident,
			     NXP_ELEMENT_VARIABLE(b)->ident, context)) {
        /* Polynomial in one variable */

	if (op == NXP_OP_PLUS) {
	    NxpElement *two;

	    vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(1, NULL));
	    _element_vector_index(vector, 0) = nxp_element_int_new(1);

	    two = nxp_element_int_new(2);
	    mon = nxp_monomial_new(two, NXP_ELEMENT(vector));
	    NXP_ELEMENT_UNREF(two);

	    args = nxp_element_list_new_simple(a);
	    el = nxp_element_mpoly_new_simple(mon, args);
	}
	else
	    return nxp_element_int_new(0);
    }
    else {
	NxpElement *one, *one1;

	/* Multivariable polynomial */
	vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(2, NULL));
	one = nxp_element_int_new(1);
	one1 = nxp_element_int_new((op == NXP_OP_PLUS) ? 1 : -1);
	_element_vector_set(vector, 0, one);
	_element_vector_index(vector, 1) = nxp_element_int_new(0);

	mon1 = nxp_monomial_new(one, NXP_ELEMENT(vector));
	NXP_ELEMENT_UNREF(vector);

	vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(2, NULL));
	_element_vector_index(vector, 0) = nxp_element_int_new(0);
	_element_vector_set(vector, 1, one);
	mon2 = nxp_monomial_new(one1, NXP_ELEMENT(vector));

	args = nxp_element_list_new_simple(a);
	nxp_element_list_append(args, b);
	mon = nxp_element_list_new_simple(mon1);
	NXP_ELEMENT_UNREF(mon1);
	nxp_element_list_append(mon, mon2);
	NXP_ELEMENT_UNREF(mon2);
	el = nxp_element_mpoly_new(mon, args);
	NXP_ELEMENT_UNREF(one);
	NXP_ELEMENT_UNREF(one1);
    }
    NXP_ELEMENT_UNREF(args);
    NXP_ELEMENT_UNREF(mon);
    NXP_ELEMENT_UNREF(vector);
    return el;
}

static NxpElement* simp_eq2(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error) {
    if (nxp_identifier_equal(NXP_ELEMENT_VARIABLE(a)->ident,
			     NXP_ELEMENT_VARIABLE(b)->ident, context)) {
	return nxp_element_int_new(1);
    } else
	return nxp_element_int_new(0);
}

static void destroy_func(NxpElement *elem)
{
    nxp_identifier_unref(NXP_ELEMENT_VARIABLE(elem)->ident);
}

#define BIND_LEFT(eval,func,op)    nxp_bind_operator_function (op,	\
				0, NXP_ELEMENT_TYPE_VARIABLE,	\
  				eval,				\
  				func,				\
  				0, 				\
				NULL); 
#define BIND_RIGHT(eval,func,op)    nxp_bind_operator_function (op,	\
				NXP_ELEMENT_TYPE_VARIABLE, 0,	\
  				eval,				\
  				func,				\
  				0, 				\
				NULL); 
#define BIND(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_VARIABLE, \
							 NXP_ELEMENT_TYPE_VARIABLE, \
    			                                 eval, symp, 0, NULL);
int nxp_element_variable_get_type(void)
{
    static int nxp_element_type_variable = 0;

    if (nxp_element_type_variable == 0) {
	nxp_element_type_variable = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
		nxp_element_type_variable, /* type to declare */
		sizeof(NxpElementTypeInfo), /* size of our typeinfo */
		0, /* types from that it comes : here, it's nothing */
		sizeof(NxpElementTypeInfo)); /* size of the parent typeinfo */
    }
    return nxp_element_type_variable;
}

void nxp_element_variable_type_init(void)
{
    typeinfo = nxp_element_type_get_info(nxp_element_variable_get_type());
    typeinfo->name = g_strdup("variable");
    typeinfo->derive_func = nxp_variable_derive;
    typeinfo->destroy_func = destroy_func;
    typeinfo->eval_func = eval_func;
    typeinfo->simplify_func = simplify_func;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = str_func;
    typeinfo->truth_func = truth_func;
    typeinfo->dependson_func = dependson_func;
    typeinfo->substitute_func = substitute_func;

    nxp_bind_operator_function(NXP_OP_EQ, NXP_ELEMENT_TYPE_VARIABLE, NXP_ELEMENT_TYPE_ANY,
			       eval_assign, simp_assign, 
			       0, NULL);
    nxp_bind_operator_function(NXP_OP_COLON_EQ, NXP_ELEMENT_TYPE_VARIABLE, NXP_ELEMENT_TYPE_ANY,
			       eval_store, eval_store, 
			       0, NULL);
    BIND(NULL, simp_eq2, NXP_OP_EQ2);
    BIND(NULL, simp_add, NXP_OP_PLUS);
    BIND(NULL, simp_add, NXP_OP_MINUS);
    BIND(NULL, simp_mul, NXP_OP_STAR);
    BIND(NULL, simp_div, NXP_OP_SLASH);
    BIND_LEFT(minus_left_eval, minus_left, NXP_OP_MINUS_LEFT);
    BIND_LEFT(inc_eval, inc_simp, NXP_OP_PLUS2_LEFT);
    BIND_LEFT(dec_eval, dec_simp, NXP_OP_MINUS2_LEFT);
}
