/* -*- c -*- */
#include <gsl/gsl_vector.h>
#include <gsl/gsl_histogram.h>
#include <gsl/gsl_histogram2d.h>
#include "kernel/types/types_misc.h"
#include "nxp_kernel.h"
#include "gsl.h"
#include "i18n.h"
#include "gslhist-prototypes.generated.h"

/*
 * The variable to handle the histogram type information 
 */
static NxpElementTypeInfo *histogram_typeinfo = NULL;
static NxpElementTypeInfo *histogram2d_typeinfo = NULL;

static void destroy_pdf(NxpElement *elem) {
    if (NXP_ELEMENT_HISTOGRAM(elem)->pdf)
	gsl_histogram_pdf_free(NXP_ELEMENT_HISTOGRAM(elem)->pdf);
}

NxpElement *nxp_gsl_histogram_new(gsl_histogram * histogram)
{
    NxpElement *el;
    el = nxp_element_alloc(histogram_typeinfo,
			   sizeof(NxpElementHistogram));
    NXP_ELEMENT_HISTOGRAM(el)->histogram = histogram;
    NXP_ELEMENT_HISTOGRAM(el)->pdf = NULL;
    return el;
}

NxpElement *nxp_gsl_histogram_dup(NxpElementHistogram *histogram) {
    return nxp_gsl_histogram_new(gsl_histogram_clone(histogram->histogram));
}


NxpElement *nxp_gsl_histogram2d_new(gsl_histogram2d* histogram)
{
    NxpElement *el;
    el = nxp_element_alloc(histogram2d_typeinfo,
			   sizeof(NxpElementHistogram2d));
    NXP_ELEMENT_HISTOGRAM2D(el)->histogram = histogram;
    return el;
}

NxpElement *nxp_gsl_histogram2d_dup(NxpElementHistogram2d *histogram) {
    return nxp_gsl_histogram2d_new(gsl_histogram2d_clone(histogram->histogram));
}


/* 
 * * ******************************************* *
 */


static NxpElement *
nxp_gsl_new_histogram(NxpElementVector * vec, ETC)
{
    GError *err = NULL;
    gsl_histogram *histogram;
    gint size;
    double *vecd;

    size = NXP_ELEMENT_VECTOR_SIZE(vec) - 1; /* ranges = bucks + 1 */
    histogram = gsl_histogram_alloc(size);
    if (!histogram) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, 
		    _("gsl cannot allocate histogram"));
	return NULL;
    }
    
    /* Evaluate */
    vecd = nxp_vector_to_double_buffer(vec, context, &err);

    /* TODO:
     *
     * vecd must be monotonically increasing!!
     * */

    if (err) {
	g_propagate_error(error, err);
	g_free(vecd);
	return NULL;
    }
    /* set ranges */
    gsl_histogram_set_ranges(histogram, vecd, size+1);
    g_free(vecd);
    return nxp_gsl_histogram_new(histogram);
}


static NxpElement *
nxp_gsl_new_histogram2d(NxpElementVector * vA,
			NxpElementVector * vB, ETC)
{
    GError *err = NULL;
    gsl_histogram2d *histogram;
    gint sizeA, sizeB;
    double *vecA,*vecB;

    sizeA = NXP_ELEMENT_VECTOR_SIZE(vA) - 1;
    sizeB = NXP_ELEMENT_VECTOR_SIZE(vB) - 1;

    histogram = gsl_histogram2d_alloc(sizeA,sizeB);
    if (!histogram) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, 
		    _("gsl cannot allocate histogram"));
	return NULL;
    }
    
    /* Evaluate */
    /* TODO: vecs must be monotonically increasing!! */
    vecA = nxp_vector_to_double_buffer(vA, context, &err); 
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    vecB = nxp_vector_to_double_buffer(vB, context, &err); 
    if (err) {
	g_free(vecA);
	g_propagate_error(error, err);
	return NULL;
    }

    /* set ranges */
    gsl_histogram2d_set_ranges(histogram, vecA, sizeA+1, vecB, sizeB+1);
    g_free(vecA);
    g_free(vecB);
    return nxp_gsl_histogram2d_new(histogram);
}

static NxpElement *
nxp_gsl_new_histogram22d(NxpElementInt *n_x,
			 NxpElementReal *l_x,
			 NxpElementReal *h_x,
			 NxpElementInt *n_y,
			 NxpElementReal *l_y,
			 NxpElementReal *h_y, ETC) {
    double lx,ly,hx,hy,tmp;
    nxp_int_t nx, ny;
    gsl_histogram2d *histogram;
    GError *err = NULL;

    nx = nxp_element_int_get_si(n_x, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    ny = nxp_element_int_get_si(n_y, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    histogram = gsl_histogram2d_alloc(nx, ny);
    if (!histogram) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, 
		    _("gsl cannot allocate histogram"));
	return NULL;
    }
    lx = nxp_element_r_get_value_as_double(l_x);
    ly = nxp_element_r_get_value_as_double(l_y);
    hx = nxp_element_r_get_value_as_double(h_x);
    hy = nxp_element_r_get_value_as_double(h_y);
    if (lx>hx) { tmp=lx;lx=hx;hx=tmp; }
    if (ly>hy) { tmp=ly;ly=hy;hy=tmp; }

    gsl_histogram2d_set_ranges_uniform(histogram,
				       lx,hx,ly,hy);
    return nxp_gsl_histogram2d_new(histogram);
}

static NxpElement *
nxp_gsl_new_histogram2(NxpElementInt *nr, 
		       NxpElementReal *low,
		       NxpElementReal *high, ETC)
{
    gsl_histogram *histogram;
    double l,h,s;

    histogram = gsl_histogram_alloc(nxp_element_r_get_value_as_int(nr));
    if (!histogram) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, 
		    _("gsl cannot allocate histogram"));
	return NULL;
    }
    
    /* verify limits */
    l = nxp_element_r_get_value_as_double(low);
    h = nxp_element_r_get_value_as_double(high);
    if (l>h) { s=l;l=h;h=s; }

    /* set ranges */
    gsl_histogram_set_ranges_uniform(histogram, l, h);
    return nxp_gsl_histogram_new(histogram);
}


static NxpElement *
nxp_gsl_histogram_increment(NxpElementHistogram *histogram,
			    NxpElementReal *bin, ETC)
{
    NxpElement *new_elem = nxp_gsl_histogram_dup(histogram);
    gsl_histogram_increment(NXP_ELEMENT_HISTOGRAM(new_elem)->histogram,
			    nxp_element_r_get_value_as_double(bin));
    return new_elem;
}

static NxpElement *
nxp_gsl_histogram2d_increment(NxpElementHistogram2d *histogram,
			    NxpElementReal *binX,
			    NxpElementReal *binY, ETC)
{
    NxpElement *new_elem = nxp_gsl_histogram2d_dup(histogram);
    gsl_histogram2d_increment(NXP_ELEMENT_HISTOGRAM2D(new_elem)->histogram,
			      nxp_element_r_get_value_as_double(binX), nxp_element_r_get_value_as_double(binY));
    return new_elem;
}



static NxpElement *
nxp_gsl_histogram_accumulate(NxpElementHistogram *histogram,
			     NxpElementReal *bin,
			     NxpElementReal *weight, ETC)
{
    NxpElement *h = nxp_gsl_histogram_dup(histogram);
    gsl_histogram_accumulate(NXP_ELEMENT_HISTOGRAM(h)->histogram,
			     nxp_element_r_get_value_as_double(bin),
			     nxp_element_r_get_value_as_double(weight));
    return h;
}

static NxpElement *
nxp_gsl_histogram2d_accumulate(NxpElementHistogram2d *histogram,
			     NxpElementReal *Xbin,
			     NxpElementReal *Ybin,
			     NxpElementReal *weight, ETC)
{
    NxpElement *h = nxp_gsl_histogram2d_dup(histogram);
    gsl_histogram2d_accumulate(NXP_ELEMENT_HISTOGRAM2D(h)->histogram,
			       nxp_element_r_get_value_as_double(Xbin),
			       nxp_element_r_get_value_as_double(Ybin),
			       nxp_element_r_get_value_as_double(weight));
    return h;
}


static NxpElement *
nxp_gsl_histogram_find(NxpElementHistogram *histogram,
		       NxpElementReal *x, ETC) {
    int i;
    if (gsl_histogram_find(histogram->histogram, nxp_element_r_get_value_as_double(x), &i) == GSL_SUCCESS) {
	return nxp_element_int_new(i);
    } else {
	return NULL;
    }
}

static NxpElement *
nxp_gsl_histogram2d_find(NxpElementHistogram2d *histogram,
			 NxpElementReal *x,
			 NxpElementReal *y, ETC) {
    int i,j;
    if (gsl_histogram2d_find(histogram->histogram, nxp_element_r_get_value_as_double(x), nxp_element_r_get_value_as_double(y), &i, &j) == GSL_SUCCESS) {
	return nxp_pair_new(nxp_element_int_new(i),
			    nxp_element_int_new(j));
    } else {
	return NULL;
    }
}

static NxpElement *
nxp_gsl_histogram_get(NxpElementHistogram *histogram,
		      NxpElementInt *bin, ETC) {
    NxpElement *el;
    el = nxp_element_real_new(gsl_histogram_get(histogram->histogram, nxp_element_r_get_value_as_int(bin)));
    return el;
}

static NxpElement *
nxp_gsl_histogram2d_get(NxpElementHistogram2d *histogram,
			NxpElementInt *Xbin, 
			NxpElementInt *Ybin, ETC) {
    NxpElement *el;
    el = nxp_element_real_new(gsl_histogram2d_get(histogram->histogram,
						  nxp_element_r_get_value_as_int(Xbin),
						  nxp_element_r_get_value_as_int(Ybin)));
    return el;
}

static NxpElement *
nxp_gsl_histogram_get_range(NxpElementHistogram *histogram,
			    NxpElementInt *bin, ETC) {
    double L, H;
    if (gsl_histogram_get_range(histogram->histogram, nxp_element_r_get_value_as_int(bin), &L, &H)) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("Out of range"));
	return NULL;
    }
    return nxp_pair_new(nxp_element_real_new(L),
			nxp_element_real_new(H));
}

static NxpElement *
nxp_gsl_histogram2d_get_xrange(NxpElementHistogram2d *histogram,
			       NxpElementInt *bin, ETC) {
    double L, H;
    if (gsl_histogram2d_get_xrange(histogram->histogram, nxp_element_r_get_value_as_int(bin), &L, &H)) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("Out of range"));
	return NULL;
    }
    return nxp_pair_new(nxp_element_real_new(L),
			nxp_element_real_new(H));
}

static NxpElement *
nxp_gsl_histogram2d_get_yrange(NxpElementHistogram2d *histogram,
			       NxpElementInt *bin, ETC) {
    double L, H;
    if (gsl_histogram2d_get_yrange(histogram->histogram, nxp_element_r_get_value_as_int(bin), &L, &H)) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("Out of range"));
	return NULL;
    }
    return nxp_pair_new(nxp_element_real_new(L),nxp_element_real_new(H));
}

static NxpElement *
nxp_gsl_histogram_reset(NxpElementHistogram *histogram, ETC) 
{
    NxpElement *h = nxp_gsl_histogram_dup(histogram);
    gsl_histogram_reset(NXP_ELEMENT_HISTOGRAM(h)->histogram);
    return h;
}


static NxpElement *
nxp_gsl_histogram2d_reset(NxpElementHistogram2d *histogram, ETC) 
{
    NxpElement *h = nxp_gsl_histogram2d_dup(histogram);
    gsl_histogram2d_reset(NXP_ELEMENT_HISTOGRAM2D(h)->histogram);
    return h;
}


static NxpElement *
nxp_gsl_histogram_set_values(NxpElementHistogram *histogram,
			     NxpElementVector *values, ETC) {
    double *vecd;
    NxpElement *h;
    GError *err = NULL;

    vecd = nxp_vector_to_double_buffer(values, context, &err);
    if (err) { g_propagate_error(error, err); g_free(vecd); return NULL; }

    h = nxp_gsl_histogram_dup(histogram);
    if (NXP_ELEMENT_HISTOGRAM(h)->histogram->bin)
	g_free(NXP_ELEMENT_HISTOGRAM(h)->histogram->bin);
    NXP_ELEMENT_HISTOGRAM(h)->histogram->bin = vecd;
    return h;
}
		      
/**
 * Representation & str functions
 */
static GString *
repr(NxpElement * el, int *precedence)
{
    GString        *ans;
    double          low, high;
    int             i;
    size_t          size;
    if (precedence)
	*precedence = 0;
    ans = g_string_new("::gsl::histogram::newWithValues({");

    low = gsl_histogram_min(NXP_ELEMENT_HISTOGRAM(el)->histogram);
    g_string_append_printf(ans, "%lf", low);

    size = gsl_histogram_bins(NXP_ELEMENT_HISTOGRAM(el)->histogram);
    for (i = 0; i < size; i++) {
	gsl_histogram_get_range(NXP_ELEMENT_HISTOGRAM(el)->histogram,
				i, &low, &high);
	g_string_append_printf(ans, ", %lf",high);
    }
    ans = g_string_append(ans, "}, {");

    for (i = 0; i< size; i++) {
	g_string_append_printf(ans, "%lf", gsl_histogram_get(NXP_ELEMENT_HISTOGRAM(el)->histogram, i));
	if (i+1<size) {
	    ans = g_string_append_c(ans, ',');
	}
    }
    ans = g_string_append(ans, "})");
    return ans;
}

static NxpElement *nxp_gsl_histogram_max(NxpElementHistogram *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram_max(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram_min(NxpElementHistogram *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram_min(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *nxp_gsl_histogram2d_xmax(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_xmax(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_xmin(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_xmin(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_ymax(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_ymax(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_ymin(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_ymin(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *nxp_gsl_histogram_bins(NxpElementHistogram *arg1, ETC)
{
	nxp_int_t ret;
	ret = gsl_histogram_bins(arg1->histogram);
	GSL_RETURN(nxp_element_int_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_nx(NxpElementHistogram2d *arg1, ETC)
{
	nxp_int_t ret;
	ret = gsl_histogram2d_nx(arg1->histogram);
	GSL_RETURN(nxp_element_int_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_ny(NxpElementHistogram2d *arg1, ETC)
{
	nxp_int_t ret;
	ret = gsl_histogram2d_ny(arg1->histogram);
	GSL_RETURN(nxp_element_int_new(ret));
}


static NxpElement *nxp_gsl_histogram_max_val(NxpElementHistogram *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram_max_val(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_max_val(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_max_val(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *nxp_gsl_histogram_max_bin(NxpElementHistogram *arg1, ETC)
{
	nxp_int_t ret;
	ret = gsl_histogram_max_bin(arg1->histogram);
	GSL_RETURN(nxp_element_int_new(ret));
}

static NxpElement *
nxp_gsl_histogram2d_max_bin(NxpElementHistogram2d *histogram, ETC) {
    size_t i,j;
    gsl_histogram2d_max_bin(histogram->histogram, &i, &j);
    return nxp_pair_new(nxp_element_int_new(i),
			nxp_element_int_new(j));
}

static NxpElement *nxp_gsl_histogram_min_val(NxpElementHistogram *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram_min_val(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_min_val(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_min_val(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *nxp_gsl_histogram_min_bin(NxpElementHistogram *arg1, ETC)
{
	nxp_int_t ret;
	ret = gsl_histogram_min_bin(arg1->histogram);
	GSL_RETURN(nxp_element_int_new(ret));
}

static NxpElement *
nxp_gsl_histogram2d_min_bin(NxpElementHistogram2d *histogram, ETC) {
    size_t i,j;
    gsl_histogram2d_min_bin(histogram->histogram, &i, &j);
    return nxp_pair_new(nxp_element_int_new(i),
			nxp_element_int_new(j));
}

static NxpElement *nxp_gsl_histogram_mean(NxpElementHistogram *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram_mean(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_xmean(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_xmean(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_ymean(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_ymean(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *nxp_gsl_histogram_sigma(NxpElementHistogram *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram_sigma(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_xsigma(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_xsigma(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_ysigma(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_ysigma(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *nxp_gsl_histogram_sum(NxpElementHistogram *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram_sum(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_histogram2d_sum(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_sum(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *nxp_gsl_histogram2d_cov(NxpElementHistogram2d *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_histogram2d_cov(arg1->histogram);
	GSL_RETURN(nxp_element_real_new(ret));
}


static NxpElement *
nxp_gsl_histogram_scale(NxpElementHistogram *histogram,
			NxpElementReal *scale, ETC) {
    NxpElement *h = nxp_gsl_histogram_dup(histogram);
    gsl_histogram_scale(NXP_ELEMENT_HISTOGRAM(h)->histogram, nxp_element_r_get_value_as_double(scale));
    return h;
}

static NxpElement *
nxp_gsl_histogram2d_scale(NxpElementHistogram2d *histogram,
			  NxpElementReal *scale, ETC) {
    NxpElement *h = nxp_gsl_histogram2d_dup(histogram);
    gsl_histogram2d_scale(NXP_ELEMENT_HISTOGRAM2D(h)->histogram, nxp_element_r_get_value_as_double(scale));
    return h;
}

static NxpElement *
nxp_gsl_histogram_shift(NxpElementHistogram *histogram,
			NxpElementReal *offset, ETC) {
    NxpElement *h = nxp_gsl_histogram_dup(histogram);
    gsl_histogram_shift(NXP_ELEMENT_HISTOGRAM(h)->histogram, nxp_element_r_get_value_as_double(offset));
    return h;
}

static NxpElement *
nxp_gsl_histogram2d_shift(NxpElementHistogram2d *histogram,
			  NxpElementReal *offset, ETC) {
    NxpElement *h = nxp_gsl_histogram2d_dup(histogram);
    gsl_histogram2d_shift(NXP_ELEMENT_HISTOGRAM2D(h)->histogram, nxp_element_r_get_value_as_double(offset));
    return h;
}

static NxpElement *
nxp_gsl_histogram_sample(NxpElementHistogram *histogram,
			 NxpElementReal *R, ETC) {
    /* First, check if we have already a PDF somewhere */
    if (histogram->pdf == NULL) { /* no.. create it */
	histogram->pdf = gsl_histogram_pdf_alloc(gsl_histogram_bins(histogram->histogram));
	gsl_histogram_pdf_init(histogram->pdf,
			       histogram->histogram);
    }
    return nxp_element_real_new(gsl_histogram_pdf_sample(histogram->pdf, nxp_element_r_get_value_as_double(R)));
}


/* --- */
static NxpElement* equal(NxpOp op, NxpElement* a, NxpElement* b, 
			 NxpEvalContext *context, GError **error) {
    NxpElementHistogram *h_a, *h_b;
    int bool;

    h_a = NXP_ELEMENT_HISTOGRAM(a);
    h_b = NXP_ELEMENT_HISTOGRAM(b);

    bool = gsl_histogram_equal_bins_p(h_a->histogram,
				      h_b->histogram);
    return nxp_element_int_new(bool);
}

static NxpElement* equal2d(NxpOp op, NxpElement* a, NxpElement* b, 
			   NxpEvalContext *context, GError **error) {
    NxpElementHistogram2d *h_a, *h_b;
    int bool;

    h_a = NXP_ELEMENT_HISTOGRAM2D(a);
    h_b = NXP_ELEMENT_HISTOGRAM2D(b);

    bool = gsl_histogram2d_equal_bins_p(h_a->histogram,
					h_b->histogram);
    return nxp_element_int_new(bool);
}

#define D(func,gslfunc) static NxpElement* func(NxpOp op, NxpElement *a, NxpElement *b,\
		       NxpEvalContext *context, GError **error) {\
    NxpElementHistogram *h_a, *h_b;\
    NxpElement *h_c;\
    gsl_histogram *hA, *hB, *hC;\
    h_a = NXP_ELEMENT_HISTOGRAM(a);\
    h_b = NXP_ELEMENT_HISTOGRAM(b);\
    hA = h_a->histogram;\
    hB = h_b->histogram;\
    hC = gsl_histogram_clone(hA);\
    gslfunc(hC, hB);\
    h_c = nxp_gsl_histogram_new(hC);\
    return h_c;}

#define D2D(func,gslfunc) static NxpElement* func(NxpOp op, NxpElement *a, NxpElement *b,\
		       NxpEvalContext *context, GError **error) {\
    NxpElementHistogram2d *h_a, *h_b;\
    NxpElement *h_c;\
    gsl_histogram2d *hA, *hB, *hC;\
    h_a = NXP_ELEMENT_HISTOGRAM2D(a);\
    h_b = NXP_ELEMENT_HISTOGRAM2D(b);\
    hA = h_a->histogram;\
    hB = h_b->histogram;\
    hC = gsl_histogram2d_clone(hA);\
    gslfunc(hC, hB);\
    h_c = nxp_gsl_histogram2d_new(hC);\
    return h_c;}

D(sum, gsl_histogram_add)
D2D(sum2d, gsl_histogram2d_add)

D(sub, gsl_histogram_sub)
D2D(sub2d, gsl_histogram2d_sub)

D(mul, gsl_histogram_mul)
D2D(mul2d, gsl_histogram2d_mul)

D(divi,gsl_histogram_div)
D2D(div2d,gsl_histogram2d_div)

static void destroy(NxpElement *elem)
{
    gsl_histogram_free(NXP_ELEMENT_HISTOGRAM(elem)->histogram);
    destroy_pdf(elem);
}

static void destroy2d(NxpElement *elem)
{
    gsl_histogram2d_free(NXP_ELEMENT_HISTOGRAM2D(elem)->histogram);
}


static NxpElement *
nxp_gsl_new_histogram_full(NxpElementVector *ranges,
			   NxpElementVector *values, ETC) {
    NxpElement *histogram;
    GError *err = NULL;
    double *vector;

    if (NXP_ELEMENT_VECTOR_SIZE(ranges) != NXP_ELEMENT_VECTOR_SIZE(values) + 1) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT,
		    _("Values vector must be 1 element smaller than ranges vector"));
	return NULL;
    }

    histogram = nxp_gsl_new_histogram(ranges, named_args, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    
    vector = nxp_vector_to_double_buffer(values, context, &err);
    if (err) {
	NXP_ELEMENT_UNREF(histogram);
	g_propagate_error(error, err);
	return NULL;
    }
    if (NXP_ELEMENT_HISTOGRAM(histogram)->histogram->bin)
	g_free(NXP_ELEMENT_HISTOGRAM(histogram)->histogram->bin);
    NXP_ELEMENT_HISTOGRAM(histogram)->histogram->bin = vector;
    return histogram;
    
}

int nxp_element_histogram_get_type(void)
{
    static int nxp_element_type_histogram = 0;
    if (nxp_element_type_histogram == 0) {
	nxp_element_type_histogram = nxp_element_next_type();
	histogram_typeinfo = nxp_element_type_register_info(
					    nxp_element_type_histogram,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_histogram;
}

int nxp_element_histogram2d_get_type(void)
{
    static int nxp_element_type_histogram2d = 0;
    if (nxp_element_type_histogram2d == 0) {
	nxp_element_type_histogram2d = nxp_element_next_type();
	histogram2d_typeinfo = nxp_element_type_register_info(
					    nxp_element_type_histogram2d,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_histogram2d;
}

void
nxp_gsl_histogram_init(NxpEvalContext * context)
{
    NxpIdentifier  *id;

    histogram_typeinfo   = nxp_element_type_get_info(nxp_element_histogram_get_type());
    histogram2d_typeinfo = nxp_element_type_get_info(nxp_element_histogram2d_get_type());

    histogram_typeinfo->name   = g_strdup("gslHistogram");
    histogram2d_typeinfo->name = g_strdup("gslHistogram2d");

    histogram_typeinfo->eval_func     = nxp_eval_return_self;
    histogram2d_typeinfo->eval_func   = nxp_eval_return_self;

    histogram_typeinfo->simplify_func   = nxp_eval_return_self;
    histogram2d_typeinfo->simplify_func = nxp_eval_return_self;

    histogram_typeinfo->derive_func   = NULL;
    histogram2d_typeinfo->derive_func = NULL;

    histogram_typeinfo->str_func   = repr;
    //histogram2d_typeinfo->str_func = show2d;

    histogram_typeinfo->repr_func = repr;
    //histogram2d_typeinfo->repr_func = repr2d;

    histogram_typeinfo->destroy_func = destroy;
    histogram2d_typeinfo->destroy_func = destroy2d;

    gslhist_prototypes_init();

    /* BIND OPERATORS */
#define BIND_FUNC(func,op)    nxp_bind_operator_function (op, 	  \
   			NXP_ELEMENT_TYPE_HISTOGRAM, NXP_ELEMENT_TYPE_HISTOGRAM, \
  			func, func, 0, NULL);

#define BIND2D_FUNC(func,op)    nxp_bind_operator_function (op, 	  \
  				NXP_ELEMENT_TYPE_HISTOGRAM2D, NXP_ELEMENT_TYPE_HISTOGRAM2D, \
  				func, func, 0, NULL);

    BIND_FUNC(equal, NXP_OP_EQ2);
    BIND2D_FUNC(equal2d, NXP_OP_EQ2);

    BIND_FUNC(sum, NXP_OP_PLUS);
    BIND2D_FUNC(sum2d, NXP_OP_PLUS);

    BIND_FUNC(sub, NXP_OP_MINUS);
    BIND2D_FUNC(sub2d, NXP_OP_MINUS);

    BIND_FUNC(divi, NXP_OP_STAR);
    BIND2D_FUNC(div2d, NXP_OP_STAR);

    BIND_FUNC(mul, NXP_OP_SLASH);
    BIND2D_FUNC(mul2d, NXP_OP_SLASH);


nxp_name_space_new(context->cwd,"gsl");
nxp_name_space_new(nxp_name_space_find_child(context->cwd,"gsl"),"histogram");
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "new");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_new_histogram, NULL, nxp_gsl_new_histogram, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "new");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR_VECTOR, nxp_gsl_new_histogram2d, NULL, nxp_gsl_new_histogram2d, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "new");
 nxp_bifunc_define(id, &nxp_prototype__INT_REAL_REAL_INT_REAL_REAL, nxp_gsl_new_histogram22d, NULL, nxp_gsl_new_histogram22d, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "new");
 nxp_bifunc_define(id, &nxp_prototype__INT_REAL_REAL, nxp_gsl_new_histogram2, NULL, nxp_gsl_new_histogram2, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "increment");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_REAL, nxp_gsl_histogram_increment, NULL, nxp_gsl_histogram_increment, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "increment");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_REAL_REAL, nxp_gsl_histogram2d_increment, NULL, nxp_gsl_histogram2d_increment, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "accumulate");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_REAL_REAL, nxp_gsl_histogram_accumulate, NULL, nxp_gsl_histogram_accumulate, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "accumulate");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_REAL_REAL_REAL, nxp_gsl_histogram2d_accumulate, NULL, nxp_gsl_histogram2d_accumulate, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "find");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_REAL, nxp_gsl_histogram_find, NULL, nxp_gsl_histogram_find, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "find");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_REAL_REAL, nxp_gsl_histogram2d_find, NULL, nxp_gsl_histogram2d_find, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "get");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_INT, nxp_gsl_histogram_get, NULL, nxp_gsl_histogram_get, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "get");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_INT_INT, nxp_gsl_histogram2d_get, NULL, nxp_gsl_histogram2d_get, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "getRange");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_INT, nxp_gsl_histogram_get_range, NULL, nxp_gsl_histogram_get_range, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "getXRange");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_INT, nxp_gsl_histogram2d_get_xrange, NULL, nxp_gsl_histogram2d_get_xrange, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "getYRange");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_INT, nxp_gsl_histogram2d_get_yrange, NULL, nxp_gsl_histogram2d_get_yrange, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "reset");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_reset, NULL, nxp_gsl_histogram_reset, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "reset");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_reset, NULL, nxp_gsl_histogram2d_reset, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "setValues");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_VECTOR, nxp_gsl_histogram_set_values, NULL, nxp_gsl_histogram_set_values, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "max");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_max, NULL, nxp_gsl_histogram_max, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "min");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_min, NULL, nxp_gsl_histogram_min, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "maxX");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_xmax, NULL, nxp_gsl_histogram2d_xmax, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "minX");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_xmin, NULL, nxp_gsl_histogram2d_xmin, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "maxY");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_ymax, NULL, nxp_gsl_histogram2d_ymax, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "minY");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_ymin, NULL, nxp_gsl_histogram2d_ymin, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "bins");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_bins, NULL, nxp_gsl_histogram_bins, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "nx");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_nx, NULL, nxp_gsl_histogram2d_nx, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "ny");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_ny, NULL, nxp_gsl_histogram2d_ny, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "maxVal");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_max_val, NULL, nxp_gsl_histogram_max_val, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "maxVal");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_max_val, NULL, nxp_gsl_histogram2d_max_val, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "maxBin");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_max_bin, NULL, nxp_gsl_histogram_max_bin, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "maxBin");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_max_bin, NULL, nxp_gsl_histogram2d_max_bin, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "minVal");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_min_val, NULL, nxp_gsl_histogram_min_val, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "minVal");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_min_val, NULL, nxp_gsl_histogram2d_min_val, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "minBin");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_min_bin, NULL, nxp_gsl_histogram_min_bin, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "minBin");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_min_bin, NULL, nxp_gsl_histogram2d_min_bin, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "mean");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_mean, NULL, nxp_gsl_histogram_mean, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "meanX");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_xmean, NULL, nxp_gsl_histogram2d_xmean, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "meanY");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_ymean, NULL, nxp_gsl_histogram2d_ymean, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "sigma");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_sigma, NULL, nxp_gsl_histogram_sigma, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "sigmaX");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_xsigma, NULL, nxp_gsl_histogram2d_xsigma, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "sigmaY");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_ysigma, NULL, nxp_gsl_histogram2d_ysigma, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "sum");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM, nxp_gsl_histogram_sum, NULL, nxp_gsl_histogram_sum, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "sum");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_sum, NULL, nxp_gsl_histogram2d_sum, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "cov");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D, nxp_gsl_histogram2d_cov, NULL, nxp_gsl_histogram2d_cov, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "scale");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_REAL, nxp_gsl_histogram_scale, NULL, nxp_gsl_histogram_scale, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "scale");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_REAL, nxp_gsl_histogram2d_scale, NULL, nxp_gsl_histogram2d_scale, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "shift");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_REAL, nxp_gsl_histogram_shift, NULL, nxp_gsl_histogram_shift, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "shift");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM2D_REAL, nxp_gsl_histogram2d_shift, NULL, nxp_gsl_histogram2d_shift, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "sample");
 nxp_bifunc_define(id, &gslhist_prototype__HISTOGRAM_REAL, nxp_gsl_histogram_sample, NULL, nxp_gsl_histogram_sample, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","histogram", "newWithValues");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR_VECTOR, nxp_gsl_new_histogram_full, NULL, nxp_gsl_new_histogram_full, context, NULL); nxp_identifier_unref(id);
}
