/* -*- c -*- */
#include <gsl/gsl_vector.h>
#include <gsl/gsl_permute.h>
#include <gsl/gsl_permute_vector.h>
#include "kernel/types/types_misc.h"
#include "kernel/types/number.h"
#include "nxp_kernel.h"
#include "gsl.h"
#include "gslpermut-prototypes.generated.h"
#include "i18n.h"

/*
 * define a macro to make value access faster (NOTE: this is not
 * implementation independent and can be changed on future gsl releases) 
 */
#define PERMUTATION_SET(p,i,val) ((p)->data[(i)]=(val))

#define PERMUTATION_ALLOC(el,size) el = nxp_element_alloc(permutation_typeinfo, sizeof(NxpElementPermutation)); NXP_ELEMENT_PERMUTATION(el)->data = gsl_permutation_alloc(size);


/*
 * The variable to handle the permutation type information 
 */
static NxpElementTypeInfo *permutation_typeinfo = NULL;

/*
 * To free the memory... NOTE: This assumes ref_count = 1 and should be
 * used only on special cases 
 */
static void
nxp_gsl_delete_permutation(NxpElementPermutation * p)
{
    gsl_permutation_free(p->data);
    g_free(p);
}

/*
 * Initialize namespace 
 */

/*
 * Constructor for the identity permutation (from an integer -- the
 * permutation size) 
 */
static NxpElement *
nxp_gsl_new_permutation(NxpElementInt * size, ETC)
{
    NxpElement     *el;
    el = nxp_element_alloc(permutation_typeinfo,
			   sizeof(NxpElementPermutation));
    NXP_ELEMENT_PERMUTATION(el)->data =
	gsl_permutation_calloc(nxp_element_r_get_value_as_int(size));
    return el;
}

NxpElement     *
nxp_gsl_permutation_new(gsl_permutation * permut)
{
    NxpElement     *el;
    el = nxp_element_alloc(permutation_typeinfo,
			   sizeof(NxpElementPermutation));
    NXP_ELEMENT_PERMUTATION(el)->data = permut;
    return el;
}

/*
 * Constructor for a permutation from a vector 
 */
static NxpElement *
nxp_gsl_new_permutation_(NxpElementVector * vec, ETC)
{
    NxpElement     *el;
    int             iterator;

    PERMUTATION_ALLOC(el, NXP_ELEMENT_VECTOR_SIZE(vec));

    el = nxp_element_alloc(permutation_typeinfo,
			   sizeof(NxpElementPermutation));
    NXP_ELEMENT_PERMUTATION(el)->data =
	gsl_permutation_alloc(NXP_ELEMENT_VECTOR_SIZE(vec));
    for (iterator = 0; iterator < NXP_ELEMENT_VECTOR_SIZE(vec); iterator++) {
	NxpElement *x = NXP_ELEMENT_VECTOR_ELEM(vec, iterator);
	if (NXP_ELEMENT_IS_R(x)) {
	    PERMUTATION_SET(NXP_ELEMENT_PERMUTATION(el)->data, iterator,
			    (size_t) nxp_element_r_get_value_as_double(NXP_ELEMENT_R(x)));
	} else {
	    nxp_gsl_delete_permutation(NXP_ELEMENT_PERMUTATION(el));
	    g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
			_("Permutation elements must be integers"));
	    return NULL;
	}
    }
    if (!gsl_permutation_valid(NXP_ELEMENT_PERMUTATION(el)->data)) {
	return el;
    } else {
	nxp_gsl_delete_permutation(NXP_ELEMENT_PERMUTATION(el));
	g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
		    _("Invalid permutation data"));
	return NULL;
    }
}

static NxpElement *
nxp_gsl_dup(NxpElementPermutation * p)
{
    NxpElement     *el;
    int             i,
                    size = gsl_permutation_size(p->data);
    PERMUTATION_ALLOC(el, size);
    for (i = 0; i < size; ++i)
	PERMUTATION_SET(NXP_ELEMENT_PERMUTATION(el)->data, i,
			gsl_permutation_get(p->data, i));
    return el;
}


static NxpElement *
nxp_gsl_permutation_swap(NxpElementPermutation * p,
			 NxpElementInt * i, NxpElementInt * j, ETC)
{
    NxpElement     *el;
    el = nxp_gsl_dup(p);
    gsl_permutation_swap(NXP_ELEMENT_PERMUTATION(el)->data, nxp_element_r_get_value_as_int(i),
			 nxp_element_r_get_value_as_int(j));
    GSL_RETURN(el);
}

static NxpElement *
nxp_gsl_permutation_size(NxpElementPermutation * p, ETC)
{
    NxpElement     *el =
	nxp_element_real_new(gsl_permutation_size(p->data));
    GSL_RETURN(el);
}

static NxpElement *
nxp_gsl_permutation_reverse(NxpElementPermutation * p, ETC)
{
    NxpElement     *el = nxp_gsl_dup(p);
    gsl_permutation_reverse(NXP_ELEMENT_PERMUTATION(el)->data);
    GSL_RETURN(el);
}

static NxpElement *
nxp_gsl_permutation_next(NxpElementPermutation * p, ETC)
{
    NxpElement     *el = nxp_gsl_dup(p);
    gsl_permutation_next(NXP_ELEMENT_PERMUTATION(el)->data);
    GSL_RETURN(el);
}

static NxpElement *
nxp_gsl_permutation_prev(NxpElementPermutation * p, ETC)
{
    NxpElement     *el = nxp_gsl_dup(p);
    gsl_permutation_prev(NXP_ELEMENT_PERMUTATION(el)->data);
    GSL_RETURN(el);
}

/**
*** Applying Permutations
**/
static NxpElement *
nxp_gsl_permute(NxpElementPermutation * p,
		NxpElementVector * v, NxpElementInt * i, ETC)
{
    GError *err = NULL;
    int size;
    // First, check if the sizes matches
    if (gsl_permutation_size(p->data) == (size = v->size)) {
	double *vector = nxp_vector_to_double_buffer(v, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	gsl_permute(p->data->data, vector, nxp_element_r_get_value_as_int(i), size);
	GSL_RETURN(nxp_double_buffer_to_vector(vector, size));
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
		    _("Permutation and vector must have same size"));
	return NULL;
    }
}

static NxpElement *
nxp_gsl_permutation_get(NxpElementPermutation * p, NxpElementInt * i, ETC)
{
    NxpElement     *el =
	nxp_element_int_new(gsl_permutation_get(p->data, nxp_element_r_get_value_as_int(i)));
    GSL_RETURN(el);
}

static NxpElement *
nxp_gsl_permutation_inverse(NxpElementPermutation * p, ETC)
{
    NxpElement     *el;
    el = nxp_element_alloc(permutation_typeinfo,
			   sizeof(NxpElementPermutation));
    NXP_ELEMENT_PERMUTATION(el)->data =
	gsl_permutation_calloc(gsl_permutation_size(p->data));

    // gsl_permutation_inverse(INV, P);
    gsl_permutation_inverse(NXP_ELEMENT_PERMUTATION(el)->data, p->data);

    GSL_RETURN(el);
}

static NxpElement *
nxp_gsl_permute_inverse(NxpElementPermutation * p,
			NxpElementVector * v, NxpElementInt * i, ETC)
{
    GError *err = NULL;
    int size;
    /* First, check if the sizes matches */
    if (gsl_permutation_size(p->data) == (size = v->size)) {
	double *vector = nxp_vector_to_double_buffer(v, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	gsl_permute_inverse(p->data->data, vector, nxp_element_r_get_value_as_int(i), size);
	GSL_RETURN(nxp_double_buffer_to_vector(vector, size));
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
		    _("Permutation and vector must have same size"));
	return NULL;
    }
}

static NxpElement *
nxp_gsl_permute_vector(NxpElementPermutation * p,
		       NxpElementVector * v, ETC)
{
    GError *err = NULL;
    int size;
    /* First, check if the sizes matches */
    if (gsl_permutation_size(p->data) == (size = v->size)) {
	gsl_vector *vv = nxp_vector_to_gsl_vector(v, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	gsl_permute_vector(p->data, vv);
	GSL_RETURN(gsl_vector_to_nxp_vector(vv));
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
		    _("Permutation and vector must have same size"));
	return NULL;
    }
}


static NxpElement *
nxp_gsl_permute_vector_inverse(NxpElementPermutation * p,
			       NxpElementVector * v, ETC)
{
    GError *err = NULL;
    int size;
    /* First, check if the sizes matches */
    if (gsl_permutation_size(p->data) == (size = v->size)) {
	gsl_vector *vv = nxp_vector_to_gsl_vector(v, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	gsl_permute_vector_inverse(p->data, vv);
	GSL_RETURN(gsl_vector_to_nxp_vector(vv));
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
		    _("Permutation and vector must have same size"));
	return NULL;
    }
}


/**
 * Show function
 */
static GString *
show(NxpElement * el, int *precedence)
{
    GString        *ans;
    int             i;
    size_t          size;
    if (precedence)
	*precedence = 0;
    ans = g_string_new("Permutation(");
    size = gsl_permutation_size(NXP_ELEMENT_PERMUTATION(el)->data);
    for (i = 0; i < size; i++) {
	g_string_append_printf(ans, "%d",
			       gsl_permutation_get(NXP_ELEMENT_PERMUTATION
						   (el)->data, i));
	if (i < size - 1)
	    ans = g_string_append(ans, ", ");
    }
    ans = g_string_append_c(ans, ')');
    return ans;
}

/**
 * Representation function
 */
static GString *
repr(NxpElement * el, int *precedence)
{
    GString        *ans;
    int             i;
    size_t          size;
    if (precedence)
	*precedence = 0;
    ans = g_string_new("::gsl::permutation::new({");
    size = gsl_permutation_size(NXP_ELEMENT_PERMUTATION(el)->data);
    for (i = 0; i < size; i++) {
	g_string_append_printf(ans, "%d",
			       gsl_permutation_get(NXP_ELEMENT_PERMUTATION
						   (el)->data, i));
	if (i < size - 1)
	    ans = g_string_append(ans, ", ");
    }
    ans = g_string_append(ans, "})");
    return ans;
}

static void destroy(NxpElement *elem)
{
    gsl_permutation_free(NXP_ELEMENT_PERMUTATION(elem)->data);
}

int nxp_element_permutation_get_type(void)
{
    static int nxp_element_type_permutation = 0;

    if (nxp_element_type_permutation == 0) {
	nxp_element_type_permutation = nxp_element_next_type();
	permutation_typeinfo = nxp_element_type_register_info(
					nxp_element_type_permutation,
					sizeof(NxpElementTypeInfo),
					0,
					sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_permutation;
}


/**
 *
 */
void
nxp_gsl_permutation_init(NxpEvalContext * context)
{
    NxpIdentifier  *id;

    permutation_typeinfo =
	nxp_element_type_get_info(nxp_element_permutation_get_type());
    g_print("gslPermutation : type = %d\n", nxp_element_permutation_get_type());
    permutation_typeinfo->name = g_strdup("gslPermutation");
    permutation_typeinfo->eval_func = nxp_eval_return_self;
    permutation_typeinfo->simplify_func = nxp_eval_return_self;
    permutation_typeinfo->derive_func = NULL;
    permutation_typeinfo->str_func = show;
    permutation_typeinfo->repr_func = repr;
    permutation_typeinfo->destroy_func = destroy;

    gslpermut_prototypes_init();

nxp_name_space_new(context->cwd,"gsl");
nxp_name_space_new(nxp_name_space_find_child(context->cwd,"gsl"),"permutation");
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "new");
 nxp_bifunc_define(id, &nxp_prototype__INT, nxp_gsl_new_permutation, NULL, nxp_gsl_new_permutation, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "new");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_new_permutation_, NULL, nxp_gsl_new_permutation_, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "swap");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION_INT_INT, nxp_gsl_permutation_swap, NULL, nxp_gsl_permutation_swap, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "size");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION, nxp_gsl_permutation_size, NULL, nxp_gsl_permutation_size, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "reverse");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION, nxp_gsl_permutation_reverse, NULL, nxp_gsl_permutation_reverse, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "next");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION, nxp_gsl_permutation_next, NULL, nxp_gsl_permutation_next, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "previous");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION, nxp_gsl_permutation_prev, NULL, nxp_gsl_permutation_prev, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "permute");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION_VECTOR_INT, nxp_gsl_permute, NULL, nxp_gsl_permute, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "get");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION_INT, nxp_gsl_permutation_get, NULL, nxp_gsl_permutation_get, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "inverse");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION, nxp_gsl_permutation_inverse, NULL, nxp_gsl_permutation_inverse, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "permuteInverse");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION_VECTOR_INT, nxp_gsl_permute_inverse, NULL, nxp_gsl_permute_inverse, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "permute");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION_VECTOR, nxp_gsl_permute_vector, NULL, nxp_gsl_permute_vector, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","permutation", "permuteInverse");
 nxp_bifunc_define(id, &gslpermut_prototype__PERMUTATION_VECTOR, nxp_gsl_permute_vector_inverse, NULL, nxp_gsl_permute_vector_inverse, context, NULL); nxp_identifier_unref(id);
}
