#include "vector.h"
#include "nxp_kernel.h"
#include "functions.h"
#include "message.h"
#include "matrix.h"
#include "i18n.h"

static NxpElement* nxp_stdlib_matrix_max(NxpElementMatrix* matrix, ETC) {
    int width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    int height= NXP_ELEMENT_MATRIX_HEIGHT(matrix);
    int i,j;
    if (width < 1 || height < 1) {
	return nxp_element_real_new(0);
    } else {
	nxp_real_t max = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_MATRIX_ELEM_XY(matrix, 0,0));
	for (i=0; i<width; ++i) {
	    for (j=0; j<height; ++j) {
		nxp_real_t possible_max;
		possible_max = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_MATRIX_ELEM_XY(matrix, i, j));
		if (possible_max > max) { max = possible_max; }
	    }
	}
	return nxp_element_real_new(max);
    }
}

static NxpElement* nxp_stdlib_matrix_min(NxpElementMatrix* matrix, ETC) {
    int width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    int height= NXP_ELEMENT_MATRIX_HEIGHT(matrix);
    int i,j;
    if (width < 1 || height < 1) {
	return nxp_element_real_new(0);
    } else {
	nxp_real_t min = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_MATRIX_ELEM_XY(matrix, 0,0));
	for (i=0; i<width; ++i) {
	    for (j=0; j<height; ++j) {
		nxp_real_t possible_min;
		possible_min = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_MATRIX_ELEM_XY(matrix, i, j));
		if (possible_min < min) { min = possible_min; }
	    }
	}
	return nxp_element_real_new(min);
    }
}


void nxp_stdlib_matrix_init(NxpEvalContext *context, GError **error)
{
    GError *err = NULL;

    nxp_bifunc_define_global("swap_rows", &nxp_prototype__MATRIX_INT_INT,
			     nxp_stdlib_matrix_swap_rows, NULL, nxp_stdlib_matrix_swap_rows, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("swap_cols", &nxp_prototype__MATRIX_INT_INT,
			     nxp_stdlib_matrix_swap_cols, NULL, nxp_stdlib_matrix_swap_cols, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("swap_row_col", &nxp_prototype__MATRIX_INT_INT,
			     nxp_stdlib_matrix_swap_row_col, NULL, nxp_stdlib_matrix_swap_row_col, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("transpose", &nxp_prototype__MATRIX,
			     nxp_stdlib_matrix_transpose, NULL, 
			     nxp_stdlib_matrix_transpose, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("I", &nxp_prototype__INT,
			     nxp_stdlib_matrix_I, NULL, 
			     nxp_stdlib_matrix_I, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("max", &nxp_prototype__MATRIX,
			     nxp_stdlib_matrix_max, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("min", &nxp_prototype__MATRIX,
			     nxp_stdlib_matrix_min, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}
}


NxpElement* nxp_stdlib_matrix_transpose(NxpElementMatrix  *matrix, ETC) {
    gint width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    gint height = NXP_ELEMENT_MATRIX_HEIGHT(matrix);
    gint new_width = height;
    gint new_height = width;
    NxpElement* new_matrix = nxp_element_matrix_new(new_width, new_height, NULL);
    gint row, col;

    for (row = 0; row < height; row++) {
	for (col = 0; col < width; col++)
	    // This is faster than nxp_element_matrix_real_put and, I don't
	    // know, works, while the other doesn't!
	    _element_matrix_index(NXP_ELEMENT_MATRIX(new_matrix), col, row)
		= NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_YX(matrix, row, col));
    }
    return new_matrix;
}


NxpElement* nxp_stdlib_matrix_swap_row_col(NxpElementMatrix *matrix,
					   NxpElementInt *row,
					   NxpElementInt *col, ETC) {
    gint width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    gint height = NXP_ELEMENT_MATRIX_HEIGHT(matrix);
    GError *err = NULL;

    if (width == height) {
	gint the_row = nxp_element_int_get_si(NXP_ELEMENT(row), &err) - 1;
	gint the_col = nxp_element_int_get_si(NXP_ELEMENT(col), &err) - 1;
	gint row, col;
	NxpElement* new_matrix;
	gint rrow, rcol;

	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	if (the_row < 0 || the_row >= height ||
	    the_col < 0 || the_col >= height) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_OUT_OF_RANGE, _("Out of range"));
	    return NULL;
	}
	new_matrix = nxp_element_matrix_new(width, width, NULL);
	for (row = 0; row < width; row++) {
	    for (col = 0; col < width; col++) {
		if (row == the_row || col == the_col) {
		    rrow = col; rcol = row;
		} else {
		    rrow = row; rcol = col;
		}
		// This is faster than nxp_element_matrix_real_put and, I don't
		// know, works, while the other doesn't!
		_element_matrix_index(NXP_ELEMENT_MATRIX(new_matrix), rrow, rcol) =
		    NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_YX(matrix,row, col));
	    }
	}
	return new_matrix;
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("The argument should be a square matrix"));
	return NULL;
    }
}

NxpElement* nxp_stdlib_matrix_swap_rows(NxpElementMatrix *matrix,
					NxpElementInt *rowA,
					NxpElementInt *rowB, ETC) {
    GError *err = NULL;
    gint row_i  = nxp_element_int_get_si(NXP_ELEMENT(rowA), &err) - 1;
    gint row_j  = nxp_element_int_get_si(NXP_ELEMENT(rowB), &err) - 1;
    gint width  = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    gint height = NXP_ELEMENT_MATRIX_HEIGHT(matrix);
    NxpElement* new_matrix;
    int row, col;

    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    if (row_i < 0 || row_i >= height ||
	row_j < 0 || row_j >= height) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_OUT_OF_RANGE, _("Out of range"));
	return NULL;
    }

    new_matrix = nxp_element_matrix_new(width, height, NULL);
    for (row = 0; row < height; row++) {
	gint rrow;
	if (row == row_i) rrow = row_j;
	else if (row == row_j) rrow = row_i;
	else rrow = row;
	
	for (col = 0; col < width; col++) {
	    // This is faster than nxp_element_matrix_real_put and, I don't
	    // know, works, while the other doesn't!
	    _element_matrix_index(NXP_ELEMENT_MATRIX(new_matrix), rrow, col) = 
		NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_YX(matrix, row, col));
	}
    }
    return new_matrix;
}


NxpElement* nxp_stdlib_matrix_swap_cols(NxpElementMatrix *matrix,
					NxpElementInt *colA,
					NxpElementInt *colB, ETC) {
    GError *err = NULL;
    gint col_i = nxp_element_int_get_si(NXP_ELEMENT(colA), &err) - 1;
    gint col_j = nxp_element_int_get_si(NXP_ELEMENT(colB), &err) - 1;
    gint width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    gint height = NXP_ELEMENT_MATRIX_HEIGHT(matrix);
    NxpElement* new_matrix;
    int row, col;

    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    if (col_i < 0 || col_i >= width || col_j < 0 || col_j >= width) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_OUT_OF_RANGE, _("Out of range"));
	return NULL;
    }

    new_matrix = nxp_element_matrix_new(width, height, NULL);
    for (col = 0; col < width; col++) {
	gint rcol;
	if (col == col_i) rcol = col_j;
	else if (col == col_j) rcol = col_i;
	else rcol = col;
	
	for (row = 0; row < height; row++) {
	    // This is faster than nxp_element_matrix_real_put and, I don't
	    // know, works, while the other doesn't!
	    _element_matrix_index(NXP_ELEMENT_MATRIX(new_matrix), row, col) = 
		NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_YX(matrix, row, rcol));
	}
    }
    return new_matrix;
}


NxpElement* nxp_stdlib_matrix_I(NxpElementInt *size_el, ETC)
{
    int size;
    GError *err = NULL;

    size = nxp_element_int_get_si(NXP_ELEMENT(size_el), &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }

    if (!err && size >= 1) {
	gint i,j;
	NxpElementMatrix *matrix = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(size, size, NULL));

	for (i = 0; i < size; i++)
	    for (j = 0; j < size; j++)
		_element_matrix_index(matrix, i, j) = nxp_element_int_new( i == j ? 1 : 0 );

	return NXP_ELEMENT(matrix);
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT,
		    _("Argument to I[] should be positive"));
	return  NULL;
    }
}

NxpElement* bifunc__eval_det(NxpElement_t* element, ETC)
{
    NxpElement *el, *el1;

    el = nxp_element_eval(element, context, NULL);

    if (NXP_ELEMENT_GET_TYPE(el) != NXP_ELEMENT_TYPE_MATRIX) return NULL;
    if (!NXP_ELEMENT_MATRIX_IS_SQUARE(el)) return NULL;

    el1 = nxp_element_matrix_determinant(NXP_ELEMENT_MATRIX(el),
					 context, error);
    NXP_ELEMENT_UNREF(el);
    return el1;
}


NxpElement* stdlib_eval_det(NxpElementMatrix *matrix, ETC)
{
    NxpElement *e = NULL;
    GError *err = NULL;
    e = bifunc__eval_det(NXP_ELEMENT(matrix), NULL, context, &err);
    if (err) g_propagate_error(error, err);
    return e;
}


NxpElement* bifunc_simplify_det(GSList* args, ETC)
{
    NxpElement* element;
    
    if (!(element = (NxpElement*)args->data)) return NULL;
    g_return_val_if_fail(NXP_ELEMENT_GET_TYPE(element) == NXP_ELEMENT_TYPE_MATRIX, NULL);
    return nxp_element_simplify(nxp_element_matrix_determinant(NXP_ELEMENT_MATRIX(element),
							       context, error),
				context, NULL);
}


